from __future__ import annotations

import copy
import re
from typing import TYPE_CHECKING, Final

from jett.utils import is_snake_case

if TYPE_CHECKING:
    from polars import Expr as Column
    from polars import Schema, Struct


ALLOW_VALIDATE_PATTERNS: Final[tuple[str, ...]] = (
    "non_snake_case",
    "whitespace",
)


def extract_col_with_pattern(
    schema: Schema | Struct,
    patterns: list[str],
    parent_cols: list[str] | None = None,
) -> list[str]:
    """Do recursive find the colum name that does not follow the pattern
    current supported patterns are non_snake_case and whitespace.

    Args:
        schema (Schema | Struct): A Schema or Struct Polars object.
        patterns (list[str]):
        parent_cols (list[str], default None):
    """
    from polars import Array, List, Struct

    parent_cols: list[str] = parent_cols or []

    def _validate_and_append_error(
        name: str,
        p_cols: list[str],
        e_cols: list[str],
        pt: list[str],
    ) -> list[str]:
        """Child Wrapped function for extract_pyspark_column_names_with_pattern
        validate snake case or whitespace and append error columns

        Args:
            name (str): A column name.
            p_cols: A parent columns
            e_cols: An error columns
            pt:
        """
        is_found_err: bool = False
        for pattern in pt:
            if pattern == "non_snake_case" and not is_snake_case(name):
                is_found_err = True
            elif pattern == "whitespace" and " " in name:
                is_found_err = True

        if is_found_err:
            e_cols.append(
                name if len(p_cols) == 0 else ".".join(p_cols) + f".{name}"
            )

        return e_cols

    if all(p not in ALLOW_VALIDATE_PATTERNS for p in patterns):
        raise ValueError(
            f"patterns must contain value in {ALLOW_VALIDATE_PATTERNS}"
        )

    error_cols: list[str] = []
    struct: Struct = (
        schema2struct(schema) if isinstance(schema, Schema) else Struct
    )
    for field in struct.fields:
        if isinstance(field.dtype, (Array, List)):
            if isinstance(field.dtype.inner, Struct):
                _parent_cols = copy.deepcopy(parent_cols)
                _parent_cols.append(field.name)
                error_cols = error_cols + extract_col_with_pattern(
                    schema=field.dtype.inner,
                    patterns=patterns,
                    parent_cols=_parent_cols,
                )
            else:
                error_cols: list[str] = _validate_and_append_error(
                    name=field.name,
                    p_cols=parent_cols,
                    e_cols=error_cols,
                    pt=patterns,
                )
        elif isinstance(field.dtype, Struct):
            _parent_cols = copy.deepcopy(parent_cols)
            _parent_cols.append(field.name)
            error_cols: list[str] = error_cols + extract_col_with_pattern(
                schema=field.dtype,
                patterns=patterns,
                parent_cols=_parent_cols,
            )
        else:
            error_cols: list[str] = _validate_and_append_error(
                name=field.name,
                p_cols=parent_cols,
                e_cols=error_cols,
                pt=patterns,
            )
    return error_cols


def validate_col_disallow_pattern(
    schema: Schema | Struct, patterns: list[str]
) -> None:
    """Validate columns names in dataframe (support nested schema) from the
    pattern.
    """
    if all(p not in ALLOW_VALIDATE_PATTERNS for p in patterns):
        raise ValueError(
            f"Patterns must contain value in {ALLOW_VALIDATE_PATTERNS}"
        )

    error_cols = extract_col_with_pattern(schema=schema, patterns=patterns)
    if len(error_cols) > 0:
        cols: str = ", ".join(error_cols)
        raise ValueError(
            f"Please check column naming convention (must not be {patterns!r}) "
            f"on columns: {cols}"
        )


def validate_col_allow_snake_case(schema: Schema | Struct) -> None:
    """Validate columns, allow only snake case."""
    validate_col_disallow_pattern(schema=schema, patterns=["non_snake_case"])


def schema2struct(schema: Schema) -> Struct:
    """Convert Schema object to Struct.

    Args:
        schema (Schema):

    Returns:
        Struct: A Struct object that create from the Polars Schema.
    """
    from polars import Field, Struct

    return Struct([Field(name, dtype) for name, dtype in schema.items()])


def extract_cols_selectable(
    schema: Schema | Struct, prefix: str = ""
) -> list[str]:
    """Extracts all selectable columns of given schema, support all top level
    column and nested column and array column.

    Args:
        schema (Schema | Struct): A Polars Schema object.
        prefix (str, default ''): A prefix value.

    Returns:
        list[str]: All cols like:
            ["c1", "c2.f1", "c2.f2", "c3"]

    Examples:
        Input:
        >>> from polars import Field, String, Int64, Float64
        >>> Struct(
        ...     [
        ...         Field("texts", List(String())),
        ...         Field("items", List(Struct(
        ...             [
        ...                 Field("name", String()),
        ...                 Field("price", Int64()),
        ...                 Field("detail", List(Struct(
        ...                     [
        ...                         Field("field1", String()),
        ...                         Field("field2", Float64()),
        ...                     ]
        ...                 )))
        ...             ]
        ...         )))
        ...     ]
        ... )

        Output:
        >>> [
        ...     'texts',
        ...     'texts[x]',
        ...     'items',
        ...     'items[x]',
        ...     'items[x].name',
        ...     'items[x].price',
        ...     'items[x].detail',
        ...     'items[x].detail[x]',
        ...     'items[x].detail[x].field1',
        ...     'items[x].detail[x].field2',
        ... ]
    """
    from polars import Array, List, Struct

    rs: list[str] = []
    struct: Struct = (
        schema2struct(schema) if isinstance(schema, Schema) else schema
    )
    for field in struct.fields:
        rs.append(prefix + field.name)
        if isinstance(field.dtype, Struct):
            rs.extend(
                extract_cols_selectable(field.dtype, f"{prefix}{field.name}.")
            )
        elif isinstance(field.dtype, (Array, List)):
            rs.append(f"{prefix}{field.name}[x]")
            if isinstance(field.dtype.inner, Struct):
                rs.extend(
                    extract_cols_selectable(
                        field.dtype.inner, f"{prefix}{field.name}[x]."
                    )
                )
    return rs


def extract_cols_without_array(schema: Schema) -> list[str]:
    """Extract selectable columns without array type.

        It returns only list of selectable columns that are not nested array
    type return only root array column name.

    Args:
        schema (Schema):

    Returns:
        list[str]: A list of column name that extract by selectable without
            array type.
    """
    selectable_cols: list[str] = extract_cols_selectable(schema=schema)
    nested_array_cols: list[str] = [c for c in selectable_cols if "[x]" in c]
    final_selectable_cols: list[str] = [
        c for c in selectable_cols if c not in nested_array_cols
    ]

    rs: list[str] = []
    for c in final_selectable_cols:
        is_not_parent_column: bool = True
        for fc in final_selectable_cols:
            if fc != c and fc.startswith(f"{c}."):
                is_not_parent_column: bool = False

        if is_not_parent_column:
            rs.append(c)
    return rs


def col_path(path: str) -> Column:
    """Convert a dotted path notation to polars column expression supports also
    list access

    Args:
        path (str): A column path.

    Examples:
        >>> col_path('a[5].b')

    References:
        - ISSUE - https://github.com/pola-rs/polars/issues/3123
        - FIX - https://gist.github.com/ophiry/78e6e04a8fde01e58ee289febf3bc4cc
    """
    import polars as pl

    parsed_path = re.findall(r"\.(\w+)|\[(\d+)]", f".{path}")
    expr: Column = pl.col(parsed_path[0][0])
    for field, index in parsed_path[1:]:
        if field:
            expr = expr.struct.field(field)
        elif index:
            expr = expr.list.get(int(index))
    return expr



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal

from pydantic import BaseModel, Field, SecretStr
from pydantic.functional_validators import field_validator, model_validator

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, DataFrameReader, SparkSession

from ....__types import DictData
from ....models import BasicFilter, Shape
from ...__abc import BaseSource
from ..temp_storage import TempStorage

logger = logging.getLogger("jett")


def pipe_load_all_struct_pipe_project():
    return {
        "$project": {
            "_id": 1,
            "raw_json_value": 1,
        }
    }


def pipe_load_all_struct_pipe_and_fields():
    return {"$addFields": {"raw_json_value": "$$ROOT"}}


def pipe_load_all_struct_schema():
    return {
        "fields": [
            {"metadata": {}, "name": "_id", "nullable": True, "type": "string"},
            {
                "metadata": {},
                "name": "raw_json_value",
                "nullable": True,
                "type": "string",
            },
        ],
        "type": "struct",
    }


class AllStructExtraColumn(BaseModel):
    """A Sub Configuration Model for Param load_all_struct_extra_cols of
    MongoDBModel.
    """

    col: str
    dtype: str


class MongoDB(BaseSource):
    """MongoDB Spark Source model."""

    type: Literal["mongodb"]
    protocol: Literal["mongodb", "mongodb+srv"] = "mongodb"
    host: str | list[str]
    port: int = 27017
    username: str
    password: SecretStr
    database: str
    table_name: str
    mongo_schema: dict | None = Field(None, alias="schema")
    sample_size: int = 100
    auth_source: str | None = None
    pipeline: list[dict] | None = None
    partitioner: dict | None = None
    filter: list[BasicFilter] | None = None
    sample_records: int | None = None
    load_all_struct: bool = False
    load_all_struct_extra_cols: list[AllStructExtraColumn] | None = None
    allow_save_data_to_temp: bool = True
    read_preference: str = "secondaryPreferred"
    direct_connection: bool = True

    @field_validator("database", "table_name")
    def __check_space(cls, value: Any) -> Any:
        """Validate whitespace"""
        if isinstance(value, str) and " " in value:
            raise ValueError("cannot contain whitespace")
        return value

    @field_validator("sample_records")
    def __check_range(cls, value: Any) -> Any:
        """Validate number of sample records."""
        if (
            value is not None
            and isinstance(value, int)
            and (value < 1 or value > 10000)
        ):
            raise ValueError("range is between 0 and 10000")
        return value

    @field_validator("pipeline")
    def __check_empty_mongo_pipeline(cls, value: Any) -> Any:
        """Check that mongodb pipeline cannot be empty list."""
        if value is not None and isinstance(value, list) and len(value) == 0:
            raise ValueError("pipeline cannot be empty list")
        return value

    @field_validator("load_all_struct_extra_cols")
    def __check_load_all_struct_only_first_level(cls, value: Any) -> Any:
        """Check the columns cannot contain nested column, must be first level
        column only.
        """
        if value is not None and isinstance(value, list):
            for c in value:
                if "." in c.col or "struct" in c.dtype:
                    raise ValueError(f"{c} is not first level column")
        return value

    @model_validator(mode="after")
    def __post_validate(self):
        if self.load_all_struct is True and self.pipeline is not None:
            raise ValueError("please choose either load_all_struct or pipeline")

        if isinstance(self.host, str):
            hosts = self.host.split(",")
            if len(hosts) > 1:
                self.host = hosts

        if self.direct_connection is True and isinstance(self.host, list):
            raise ValueError(
                "direct_connection not allow to be true in case multi hosts, "
                "please change direct_connection to false"
            )

        return self

    def get_uri(self) -> str:
        """Get mongodb connection uri."""
        password = self.password.get_secret_value()
        if isinstance(self.host, list):
            host = ",".join(self.host)  # for multiple endpoints
        else:
            host = f"{self.host}:{self.port}"  # for single endpoints
        uri = f"{self.protocol}://{self.username}:{password}@{host}"
        _params = []
        if self.auth_source:
            _params.append(f"authSource={self.auth_source}")
        if self.read_preference is not None:
            _params.append(f"readPreference={self.read_preference}")
        if self.direct_connection:
            _params.append(
                f"directConnection={str(self.direct_connection).lower()}"
            )

        if len(_params) > 0:
            _params = "&".join(_params)
            uri = f"{uri}/?{_params}"
        logger.info("mongo uri: %s", uri.replace(password, "******"))
        return uri

    def set_pipeline_load_all_struct(
        self,
    ) -> tuple[list[Any], dict[str, Any]] | tuple[None, None]:
        """Set mongodb pipeline and pyspark schema to load all struct."""
        if self.load_all_struct:
            logger.info("load all struct of records as json string")
            project: dict[str, Any] = pipe_load_all_struct_pipe_project()
            schema: dict[str, Any] = pipe_load_all_struct_schema()
            project_cols = project["$project"]
            columns = schema["fields"]
            if self.load_all_struct_extra_cols is not None:
                for c in self.load_all_struct_extra_cols:
                    project_cols[c.col] = 1
                    columns.append(
                        {
                            "metadata": {},
                            "name": c.col,
                            "nullable": True,
                            "type": c.dtype,
                        }
                    )

            project["$project"] = project_cols
            schema["fields"] = columns
            return [
                pipe_load_all_struct_pipe_and_fields(),
                project,
            ], schema
        return None, None

    def load(self, engine: DictData, **kwargs) -> tuple[Any, Shape]:
        """Load MongoDB Source"""
        from pyspark.sql.types import StructType

        pipeline, schema = self.set_pipeline_load_all_struct()
        spark: SparkSession = engine["spark"]
        uri: str = self.get_uri()

        reader: DataFrameReader = (
            spark.read.format("mongodb")
            .option("comment", "Tool Spark MongoDB Connector")
            .option("connection.uri", uri)
            .option("database", self.database)
            .option("collection", self.table_name)
            .option("sampleSize", self.sample_size)
        )

        _mongo_schema: dict[str, Any] = schema or self.mongo_schema
        if _mongo_schema:
            logger.info("apply schema")
            reader = reader.schema(StructType.fromJson(_mongo_schema))

        if self.partitioner:
            logger.info("apply mongo partitioner")
            for k, v in self.partitioner.items():
                logger.info(f"{k}={v}")
                reader = reader.option(k, v)

        _mongo_pipeline = pipeline or self.pipeline
        if _mongo_pipeline:
            logger.info(f"apply pipeline: {_mongo_pipeline}")
            reader = reader.option("aggregation.pipeline", self._pipeline)

        df: DataFrame = reader.load()

        # NOTE: Apply filter.
        if self.filter:
            df = df.filter(" and ".join(f.get_str_cond() for f in self.filter))

        if self.sample_records:
            logger.info(f"Source - apply limit: {self.sample_records}")
            df = df.limit(self.sample_records)

        if self.allow_save_data_to_temp:
            temp_storage = TempStorage(spark, prefix=f"source__{self.type}")
            df = temp_storage.apply(df)
            rows: int = df.count()
            logger.info(f"Source - rows: {rows}")
            shape: tuple[int, int] = (rows, len(df.columns))
            engine["temp_storage"]["source"] = temp_storage
            return df, Shape.from_tuple(shape)
        return df, Shape()

    def inlet(self) -> tuple[str, str]:
        return self.type, f"{self.database}.{self.table_name}"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Annotated, Literal, Union

from pydantic import Field

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrame, SparkSession
    from pyspark.sql.connect.session import DataFrame as DataFrameRemote
    from pyspark.sql.types import StructType

    PairCol = tuple[Column, str]
    AnyDataFrame = DataFrame | DataFrameRemote

from ....__types import DictData
from ....errors import ToolTransformError
from ....models import Context, MetricOperatorGroup, MetricOperatorOrder
from .__abc import BaseSparkTransform, is_pair_col
from .compute_metric import CalculateMinMaxOfColumns, DetectSchemaChangeWithSink
from .cryptography import GCMDecrypt
from .functions import (
    SQL,
    CleanMongoJsonStr,
    DropColumns,
    ExplodeArrayColumn,
    Expr,
    FlattenAllExceptArray,
    JsonStrToStruct,
    RenameColumns,
    RenameSnakeCase,
    Scd2,
)
from .validation import ValidateColumnDisallowSpace

logger = logging.getLogger("jett")

GroupTransform = Annotated[
    Union[RenameColumns,],
    Field(
        discriminator="op",
        description="A transform that allow to use in group operator.",
    ),
]


class Group(BaseSparkTransform):
    """Group Transform model."""

    op: Literal["group"]
    transforms: list[GroupTransform] = Field(
        description=(
            "A list of operator transform model that support for grouping "
            "apply that mean it can execute with parallelism."
        )
    )

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricOperatorGroup,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> AnyDataFrame:
        """Apply to Group operator transform.

        Args:
            df (DataFrame): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperatorOrder): A metric transform that was set from
                handler step for passing custom metric data.
            spark (SparkSession, default None): A Spark session.

        Returns:

        """
        maps: dict[str, Column] = {}
        for t in self.transforms:
            metric_order = MetricOperatorOrder(type="order", trans_op=t.op)
            metric.transforms.append(metric_order)
            try:
                output: PairCol | list[PairCol] = t.apply(
                    df, engine=engine, metric=metric_order, spark=spark
                )

                # VALIDATE: Check the result type that match on the group operator.
                if is_pair_col(output):
                    rs: dict[str, Column] = {output[1]: output[2]}
                elif isinstance(output, list) and all(
                    is_pair_col(i) for i in output
                ):
                    rs: dict[str, Column] = {p[1]: p[0] for p in output}
                else:
                    raise ToolTransformError(
                        "Group on Spark engine should return the apply"
                        "group result with a pair of Column and str only."
                    )
                maps.update(rs)
            finally:
                metric_order.finish()
        df: DataFrame = df.withColumns(maps)
        return df

    def handle_apply(
        self,
        df: AnyDataFrame,
        context: Context,
        engine: DictData,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> AnyDataFrame:
        """Handle the Group Operator Apply.

        Args:
            df (AnyDataFrame): A Spark DataFrame.
            context: (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            AnyDataFrame: A applied Spark DataFrame.
        """
        logger.info(f"👷🔧 Handle Apply Operator: {self.op!r}")
        metric = MetricOperatorGroup(type="group", trans_op=self.op)
        context["metric_operator"].append(metric)
        try:
            pre_schema: StructType = df.schema
            df: DataFrame = self.apply(
                df, engine, metric=metric, spark=spark, **kwargs
            )
            if self.cache:
                logger.warning(
                    f"🏭 Cache the DataFrame after apply operator: {self.op!r}"
                )
                df.cache()
            self.sync_schema(pre_schema, df.schema, metric=metric, spark=spark)
            return df
        finally:
            metric.finish()


Transform = Annotated[
    Union[
        GCMDecrypt,
        Expr,
        SQL,
        DropColumns,
        RenameColumns,
        RenameSnakeCase,
        ExplodeArrayColumn,
        FlattenAllExceptArray,
        Scd2,
        CleanMongoJsonStr,
        JsonStrToStruct,
        ValidateColumnDisallowSpace,
        CalculateMinMaxOfColumns,
        DetectSchemaChangeWithSink,
        Group,
    ],
    Field(
        discriminator="op",
        description="A transform registry",
    ),
]



import logging
from typing import Any, Literal

from pyarrow import Table
from pydantic import Field
from pydantic.functional_validators import field_validator

from ... import Result
from ...__types import DictData
from ...models import ColDetail, Context, MetricEngine, MetricTransform
from ..__abc import BaseEngine
from .sink import Sink
from .source import Source
from .transform import Transform

logger = logging.getLogger("jett")


class Arrow(BaseEngine):
    """Arrow Engine Model.

    This engine support multiple sink.
    """

    type: Literal["arrow"] = Field(description="An engine type.")
    source: Source = Field(description="A Source model.")
    transforms: list[Transform] = Field(
        default_factory=list, description="A list of transform operators."
    )
    sink: list[Sink] = Field(
        description="A list of Sink model.",
        default_factory=list,
    )

    @field_validator(
        "sink",
        mode="before",
        json_schema_input_type=Sink | list[Sink],
    )
    def __prepare_sink(cls, data: Any) -> Any:
        """Prepare the sink field value that should be list of Sink model."""
        return [data] if not isinstance(data, list) else data

    def execute(
        self,
        context: Context,
        engine: DictData,
        metric: MetricEngine,
    ) -> Table:
        """Execute Arrow engine."""
        logger.info("Start execute with Arrow engine.")
        df: Table = self.source.handle_load(context, engine=engine)
        df: Table = self.handle_apply(df, context, engine=engine)
        for sk in self.sink:
            sk.handle_save(df, context, engine=engine)
        return df

    def set_engine_context(self, context: Context, **kwargs) -> DictData:
        return {
            "engine": self,
        }

    def set_result(self, df: Table, context: Context) -> Result:
        return Result(
            data=[],
            columns=[
                ColDetail(name=field.name, dtype=str(field.type))
                for field in df.schema
            ],
            schema_dict={field.name: field.type for field in df.schema},
        )

    def apply(
        self,
        df: Table,
        context: Context,
        engine: DictData,
        metric: MetricTransform,
        **kwargs,
    ) -> Table:
        for op in self.transforms:
            df: Table = op.handle_apply(df, context, engine=engine)
        return df



from __future__ import annotations

import json
import logging
import time
import traceback
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import ClassVar, overload

from pydantic import BaseModel, Field, TypeAdapter, ValidationError
from typing_extensions import Self

from jett.__about__ import __version__
from jett.__types import DictData, StrOrPath
from jett.engine import Engine
from jett.errors import ToolError, ToolValidationError
from jett.models import Context, Result
from jett.utils import get_dt_now, load_yaml, substitute_env_vars, write_yaml

logger = logging.getLogger("jett")


class ToolModel(BaseModel):
    """Core Tool model that keep the parsing engine model and the original
    config data.

    Constructors:
        from_yaml: Construct Tool model with a YAML file path.
        from_dict: Construct Tool model with a dict object.

    Methods:
        fetch: Revalidate model from the current config data.
    """

    model: Engine = Field(
        description="A Tool model that already create on the registry module.",
    )
    data: DictData = Field(
        description="A configuration data that store with YAML template file."
    )
    extras: DictData = Field(
        default_factory=dict, description="An extra parameters."
    )
    created_at: datetime = Field(
        default_factory=get_dt_now,
        description=(
            "A created at of this config data if it load from YAML "
            "template file. It will use the current init time when this model "
            "created from dict."
        ),
    )

    @classmethod
    def from_yaml(
        cls,
        path: StrOrPath,
        *,
        extras: DictData | None = None,
    ) -> Self:
        """Construct Tool model with a YAML file path.

        Args:
            path (str | Path): A YAML path that want to load data.
            extras (DictData, default None): An extra parameters.
        """
        return cls.from_dict(load_yaml(path), extras=extras)

    @classmethod
    def from_dict(
        cls,
        data: DictData,
        *,
        extras: DictData | None = None,
    ) -> Self:
        """Construct Tool model with a dict object.

        Args:
            data (DictData): A dict value.
            extras (DictData, default None): An extra parameter.

        Raises:
            ToolValidationError: If model validation was failed.
        """
        try:
            _extras: DictData = extras or {}
            _data: DictData = data | _extras
            return cls.model_validate(
                {"model": _data, "data": _data, "extras": _extras}
            )
        except ValidationError as e:
            logger.exception(f"Create ToolModel from dict failed:\n\t{e}")
            raise ToolValidationError(e) from e

    def fetch(self, override: DictData | None = None) -> Self:
        """Recreate ToolModel `model` field with the current configuration data.

        Returns:
            Self: The new ToolModel instance with the current config data.
        """
        try:
            self.__dict__["model"] = TypeAdapter(Engine).validate_python(
                override or self.data
            )
        except ValidationError as e:
            logger.exception(f"Fetch model failed:\n\t{e}")
            raise ToolValidationError(e) from e


class Tool:
    """Tool object is tha main Python interface object for this
    package.

    Attributes:
        c (ToolModel): A Tool model that include the `model` and `data` fields.
        _path (str | Path, default None):
        _config (DictData, default None):

    Methods:
        refresh: Refresh the config argument with reload again from config path
            or config dict object.
    """

    @overload
    def __init__(self, *, path: StrOrPath) -> None:
        """Main construction with path of the YAML file."""

    @overload
    def __init__(self, *, config: DictData) -> None:
        """Main construction with dict of data."""

    def __init__(
        self,
        *,
        path: StrOrPath | None = None,
        config: DictData | None = None,
    ) -> None:
        """Main construction

        Args:
            path:
            config:

        Raises:
            ToolError: If path and config parameters was pass together.
            ToolError: If path and config be None value together.
        """

        # NOTE: Validate parameter with rule that should not supply both of its.
        if path and config:
            raise ToolError("Please pass only one for `path` nor `config`.")
        elif not path and not config:
            raise ToolError("Both of `path` and `config` must not be empty.")

        self._path: StrOrPath | None = path
        self._config: DictData | None = config
        # NOTE: Start pre-init.
        self.pre_init()

        self.c: ToolModel = (
            ToolModel.from_yaml(path) if path else ToolModel.from_dict(config)
        )

        # NOTE: Start post-init.
        self.post_init()

    def __str__(self) -> str:
        """Override the `__str__` dunder method."""
        return f"Tool Engine: {self.c.model.type}"

    def pre_init(self) -> None:
        """Pre initialize method for calling after this tool have initialed."""

    def post_init(self) -> None:
        """Post initialize method for calling after this tool have initialed."""

    def refresh(self) -> Self:
        """Refresh the config argument with reload again from config path or
        config dict object that pass from the initialize step.
        """
        self.c: ToolModel = (
            ToolModel.from_yaml(self._path)
            if self._path
            else ToolModel.from_dict(self._config)
        )
        return self

    def wrapped_execute(self, context: Context) -> Result:
        """Wrapped Execute with different use-case of tool that inherit this
        tool object.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
        """
        # IMPORTANT: Recreate model before start handle execution.
        #   We replace env var template before execution for secrete value.
        logger.info(
            "🔐 Passing environment variables to the Tool config data and "
            "refresh ToolModel before execute."
        )
        self.c.fetch(override=substitute_env_vars(self.c.data))
        return self.c.model.handle_execute(context=context)

    def execute(
        self,
        *,
        allow_raise: bool = False,
    ) -> Result:
        """Execute the Tool operator with it model configuration.

        Args:
            allow_raise: bool
                If set be True, it will raise the error when execution was
                failed.

        Returns:
            Result: A Result object that already return from config engine
                execution result.
        """
        ts: float = time.monotonic()
        start_date: datetime = get_dt_now()
        context: Context = {
            "author": self.c.model.author,
            "owner": self.c.model.owner,
            "parent_dir": self.c.model.parent_dir,
        }
        rs: Result = Result()
        exception: Exception | None = None
        exception_name: str | None = None
        exception_traceback: str | None = None
        run_result: bool = True
        logger.info("Start Tool Execution:")
        logger.info(f"... 🚀 Tool Version: {__version__}")
        logger.info(f"... 👷 Author: {self.c.model.author}")
        logger.info(f"... 👨‍💻 Owner: {self.c.model.owner}")
        logger.info(f"... 🕒 Start: {start_date:%Y-%m-%d %H:%M:%S}")
        try:
            rs: Result = self.wrapped_execute(context=context)
            logger.info("✅ Execute successful!!!")
        except Exception as e:
            run_result = False
            exception = e
            exception_name = e.__class__.__name__
            exception_traceback = traceback.format_exc()
            logger.exception(f"😎👌🔥 Execution catch: {exception_name}")
        finally:
            # NOTE: Start prepare metric data after engine execution end. It
            #   will catch error exception class to this metric data if the
            #   execution raise an error.
            emit_context: Context = context | {
                "run_result": run_result,
                "execution_time_ms": time.monotonic() - ts,
                "execution_start_time": start_date,
                "execution_end_time": get_dt_now(),
                "exception": exception,
                "exception_name": exception_name,
                "exception_traceback": exception_traceback,
            }
            # NOTE: Raise the raised exception class if the raise flag allow
            #   from method parameter.
            if allow_raise and exception:
                logging.error(json.dumps(emit_context, default=str, indent=1))
                raise exception

            self.c.model.emit(context=emit_context)
        return rs

    def ui(self):
        """Generate UI for this valid Tool configuration data."""


class SparkSubmitTool(Tool):
    """Spark Submit Tool."""

    yaml_filename: ClassVar[str] = "tool_conf.yml"
    python_filename: ClassVar[str] = "tool_pyspark.py"
    python_code: ClassVar[str] = dedent(
        """
        from jett import Tool
        tool = Tool(path="{file}")
        tool.execute()
        """.lstrip(
            "\n"
        )
    )

    def post_init(self) -> None:
        """Post validate model type after create ToolModel that should to be the
        `spark` or `spark-submit` types
        """
        if self.c.model.type not in ("spark", "spark-submit"):
            raise ToolValidationError(
                f"Spark Submit Tool does not support for engine type: "
                f"{self.c.model.type!r}"
            )

        # NOTE: Force update type if it is `spark`.
        self.c.model.type = "spark-submit"
        self.c.data["type"] = "spark-submit"

    def set_conf_files(self, temp_path: Path) -> None:
        """Update Tool YAML filepath in the `spark-submit --files` command line
        statement.

        Args:
            temp_path (Path): A temp path.
        """
        file: str = f"file://{(temp_path / self.yaml_filename).absolute()}"
        logger.info(f"Add config files path: {file}")
        files = self.c.data.get("files", [])
        files.append(file)
        self.c.data["files"] = files

    def set_conf_entrypoint(self, temp_path: Path) -> None:
        """Write entrypoint of program and set entrypoint for spark-submit.

        Args:
            temp_path (Path): A temp path.
        """
        filepath = f"{temp_path}/{self.yaml_filename}"
        write_yaml(filepath, data=self.c.data)

        py_filepath = f"{temp_path}/{self.python_filename}"
        logger.info(f"pyspark entrypoint file path: {py_filepath}")

        # NOTE:
        #   - for spark-submit YARN cluster mode, file will be place in working
        #     directory
        #   - for local mode, it uses the temporary path created by Python
        #     Temporary directory
        config_filepath = self.yaml_filename
        if self.c.data.get("master", "").startswith("local"):
            config_filepath: str = py_filepath

        code: str = self.python_code.format(file=config_filepath)
        with open(py_filepath, mode="w", encoding="utf-8") as f:
            f.write(code)

        self.c.data["entrypoint"] = py_filepath
        self.c.data["type"] = "spark-submit"

    def wrapped_execute(self, context: Context) -> Result:
        """Wrapped Execute that will create temp-file before start Spark submit
        process.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.

        Returns:
            Result: An empty result object.
        """
        from .engine.spark import SparkSubmit

        with TemporaryDirectory(prefix="tool-") as tmp:
            temp_path: Path = Path(tmp)
            self.set_conf_files(temp_path)
            self.set_conf_entrypoint(temp_path)

            # IMPORTANT: Recreate model before start handle execution.
            #   We replace env var template before execution for secrete value.
            self.c.fetch(override=substitute_env_vars(self.c.data))

            model: SparkSubmit = SparkSubmit.model_validate(self.c.data)
            model.submit(context=context)

        return Result()


class DbtTool(Tool):
    """DBT Tool."""

    profile: ClassVar[str] = "./profile.yml"
    model_path: ClassVar[str] = "./dbt"


class RayTool(Tool):
    """Ray Tool."""


class GxTool(Tool):
    """Greate Expectation Tool."""



from pathlib import Path
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from duckdb import DuckDBPyRelation


from pydantic import Field

from .....__types import DictData
from .....models import Shape
from ....__abc import BaseSource


class S3CSVFile(BaseSource):
    """Local file data source."""

    type: Literal["s3"]
    file_format: Literal["csv"]
    path: str
    delimiter: str = "|"
    header: bool = Field(default=True)
    sample_records: int | None = 200

    def load(
        self, engine: DictData, **kwargs
    ) -> tuple["DuckDBPyRelation", Shape]:
        """Load CSV file to DuckDB Relation object."""
        import duckdb

        file_format: str = Path(self.path).suffix
        if file_format not in (".csv",):
            raise NotImplementedError(
                f"Local file format: {file_format!r} does not support for "
                f"loading with csv type."
            )
        cursor = duckdb.connect()
        cursor.execute(
            """
            INSTALL httpfs;
            LOAD httpfs;
            SET s3_region='us-east-1';
            SET s3_access_key_id='';
            SET s3_secret_access_key='';

            CREATE TABLE data AS SELECT CAST(started_at as DATE) as started_at_date, count(ride_id) as ride_id_count
            FROM read_csv_auto('s3://confessions-of-a-data-guy/*.csv')
            GROUP BY started_at_date;

            COPY data TO 's3://confessions-of-a-data-guy/ducky-results.parquet';
            """
        )
        df = cursor.sql("SELECT * FROM data")
        return df, Shape.from_tuple(df.shape)

    def inlet(self) -> tuple[str, str]:
        return "s3", self.path



import datetime
import json
from collections.abc import Callable
from decimal import Decimal
from typing import Any

from pyspark.errors import PythonException
from pyspark.sql import Column, SparkSession
from pyspark.sql.functions import udf
from pyspark.sql.types import StringType


def clean_mongo_json_udf(
    spark: SparkSession,
    name: str | None = None,
) -> Callable[[Column | str], Column]:
    """Make the clean MongDB JSON string PySpark UDF.

    Args:
        spark (SparkSession): A Spark session.
        name (str, default None): A name for register `clean_mongo_json`
            function to the Spark session.

    Example:
        >>> from pyspark.sql.functions import col
        >>> clean_json_udf = clean_mongo_json_udf()
        >>> df = df.withColumn("clean_json", clean_json_udf(col("mongo_json")))
    """

    def clean_mongo_json(json_string: str) -> str:  # pragma: no cov
        """Clean MongoDB JSON string for PySpark UDF usage with SPARK-5063 support.

            Converts MongoDB extended JSON format to standard JSON by handling
        special MongoDB data types efficiently in a single optimized function.

        Supported MongoDB Data Types:
            - $oid → String (ObjectId)
            - $date → ISO datetime string
            - $numberLong → Integer
            - $numberDecimal → String (preserves precision)
            - $numberDouble → Float

        Args:
            json_string: MongoDB JSON string to clean

        Returns:
            Standard JSON string

        Raises:
            Exception: With detailed context for debugging
        """
        keys = frozenset(
            ["$oid", "$date", "$numberDecimal", "$numberDouble", "$numberLong"]
        )

        def parse_mongo_date(value: str | int | float) -> str:
            """Parse MongoDB date to ISO string."""
            if isinstance(value, str) and ("T" in value or "Z" in value):
                return datetime.datetime.fromisoformat(
                    value.replace("Z", "+00:00")
                ).isoformat()

            # Handle numeric timestamp (milliseconds)
            timestamp = (
                float(value) / 1000.0
                if isinstance(value, str)
                else float(value) / 1000.0
            )
            return datetime.datetime.fromtimestamp(timestamp).isoformat()

        def parse_mongo_value(mongo_dict: dict) -> Any:
            """Convert MongoDB extended JSON value to standard type."""
            key, value = next(iter(mongo_dict.items()))
            parsers = {
                "$oid": str,
                "$date": parse_mongo_date,
                "$numberLong": int,
                "$numberDecimal": lambda v: str(Decimal(str(v))),
                "$numberDouble": float,
            }
            return parsers.get(key, lambda v: v)(value)

        def process_obj(obj: Any, path: str = "") -> Any:
            """Recursively process object, converting MongoDB types."""
            if isinstance(obj, dict):
                if not obj:
                    return {}

                # NOTE: Handle MongoDB special value (single key from keys)
                if len(obj) == 1 and next(iter(obj)) in keys:
                    try:
                        return parse_mongo_value(obj)
                    except Exception as err:
                        raise PythonException(
                            f"MongoDB value parsing failed at '{path}': {err}"
                        ) from err

                # NOTE: Process regular dict
                rs = {}
                for k, v in obj.items():
                    if v is not None:  # Skip null values
                        current_path = f"{path}.{k}" if path else k
                        rs[k] = process_obj(v, current_path)
                return rs

            elif isinstance(obj, list):
                return [
                    process_obj(item, f"{path}[{i}]")
                    for i, item in enumerate(obj)
                ]
            return obj

        try:
            if not json_string or not json_string.strip():
                return "{}"

            data = json.loads(json_string)
            result = process_obj(data)
            return json.dumps(result, separators=(",", ":"))

        except json.JSONDecodeError as e:
            preview = (
                json_string[:100] + "..."
                if len(json_string) > 100
                else json_string
            )
            raise PythonException(
                f"Invalid JSON: {e}. Input: '{preview}'"
            ) from e
        except Exception as e:
            if "MongoDB value parsing failed" in str(e):
                raise  # Re-raise parsing errors with context
            preview = (
                json_string[:50] + "..."
                if len(json_string) > 50
                else json_string
            )
            raise PythonException(
                f"Processing failed: {e}. Input: '{preview}'"
            ) from e

    spark.udf.register(
        name or "clean_mongo_json", clean_mongo_json, StringType()
    )
    return udf(lambda text: clean_mongo_json(text), StringType())



import logging
from typing import Literal

from pyarrow import Table

from ....__types import DictData
from ....utils import to_snake_case
from .__abc import BaseArrowTransform

logger = logging.getLogger("jett")


class RenameSnakeCase(BaseArrowTransform):
    """Rename All Columns to Snakecase on the Arrow Table."""

    op: Literal["rename_snakecase"]

    def apply(self, df: Table, engine: DictData, **kwargs) -> Table:
        """Apply to Rename Columns to Snake case."""
        old_cols = df.column_names
        new_cols: dict[str, str] = {}
        logger.info("🔧 Start Apply Rename to Snakecase:")
        for c in old_cols:
            new_col: str = to_snake_case(c)
            logger.info(f"... {c!r} to {new_col!r}")
            new_cols[c] = new_col
        return df.rename_columns(new_cols)



from __future__ import annotations

import logging
from datetime import timedelta
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.types import StructField

from .....utils import clean_string, get_dt_now

logger = logging.getLogger("jett")


class TableMaintenance(BaseModel):
    """Maintenance Utility Model for Iceberg Table

    :param database: db name
    :param table_name: table name
    :param retention_days: a number of days that we will retain the data in the table
    :param retention_key: a timestamp column which will be used to filter data to be deleted
    :param keep_snapshot_days: a number of days that we will keep the snapshots,
        the snapshots that day older than param will be deleted
    :param keep_snapshot_at_least: a number of snapshots which will be retained after snapshot's expiration procedure
    """

    database: str = Field(description="A database name.")
    table_name: str = Field(description="A table name.")
    retention_days: int | None = None
    retention_key: str | None = None
    keep_snapshot_days: int | None = None
    keep_snapshot_at_least: int | None = None

    def validate_retention_key(self, df: DataFrame) -> None:
        """Validate retention key, it must be a timestamp column."""
        from pyspark.sql.types import TimestampType

        if self.retention_days is None:
            return

        found_col: list[StructField] = [
            col for col in df.schema if col.name == self.retention_key
        ]
        if len(found_col) == 0:
            raise ValueError(
                f"retention key: {self.retention_key} not found in DataFrame"
            )

        if not isinstance(found_col[0].dataType, TimestampType):
            raise ValueError(
                f"retention key: {self.retention_key} is not a timestamp column"
            )

    @staticmethod
    def _calculate_past_date(days: int) -> str:
        """
        calculate past datetime and result as date with hh:mm:ss of 00:00:00
        """
        current_ts = get_dt_now()
        new_ts = current_ts - timedelta(days=days)
        new_ts = new_ts.date()
        new_ts = f"{new_ts} 00:00:00"
        return new_ts

    def is_snapshot_outdated(self, spark: SparkSession) -> bool:
        """
        check is snapshot outdated
        """
        expired_at = self._calculate_past_date(days=self.keep_snapshot_days)
        sql = f"""
        SELECT COUNT(*) AS cnt FROM {self.database}.{self.table_name}.snapshots
        WHERE committed_at < '{expired_at}'
        """
        result = spark.sql(sql).collect()
        result = result[0]

        is_outdated = False
        if result.cnt > 0:
            is_outdated = True

        return is_outdated

    def expire_snapshot(self, spark: SparkSession) -> None:
        """
        expire snapshots that are older than given number of keep_snapshot_days
        """
        is_outdated = self.is_snapshot_outdated(spark=spark)
        if is_outdated:
            expired_at = self._calculate_past_date(days=self.keep_snapshot_days)
            logger.info("expire snapshots older than: %s", expired_at)
            catalog = spark.catalog.currentCatalog()
            retain_sql = f"""
            CALL {catalog}.system.expire_snapshots(
                '{self.database}.{self.table_name}',
                TIMESTAMP '{expired_at}',
                {self.keep_snapshot_at_least}
            )
            """
            retain_sql = clean_string(text=retain_sql)
            logger.info(retain_sql)
            result = spark.sql(retain_sql).collect()
            result = result[0]
            logger.info(
                "number of data files deleted: %s",
                result.deleted_data_files_count,
            )
            logger.info(
                "number of position files deleted: %s",
                result.deleted_position_delete_files_count,
            )
            logger.info(
                "number of equality files deleted: %s",
                result.deleted_equality_delete_files_count,
            )
            logger.info(
                "number of manifest files deleted: %s",
                result.deleted_manifest_files_count,
            )
            logger.info(
                "number of manifest list files deleted: %s",
                result.deleted_manifest_lists_count,
            )

    def retain_data(self, spark: SparkSession) -> None:
        """Retain data based on number of days."""
        if self.retention_days is None:
            return
        retain_at = self._calculate_past_date(days=self.retention_days)
        logger.info("delete data older than: %s", retain_at)
        delete_sql = f"""
        DELETE FROM {self.database}.{self.table_name}
        WHERE {self.retention_key} < '{retain_at}'
        """
        delete_sql = clean_string(text=delete_sql)
        logger.info(delete_sql)
        spark.sql(delete_sql)



import json
from typing import Any, Literal, TypeVar

from ...models import MetricData


class BaseConvertor:

    def __init__(self, data: MetricData, custom_metric: dict[str, Any]):
        """
        Args:
            data (MetricData): A metric data.
            custom_metric (dict[str, Any]): A custom metric data.
        """
        self.data = data
        self.custom_metric = custom_metric

    def convert(self) -> dict[str, Any]: ...


class BasicConvertor(BaseConvertor):

    def convert(self) -> dict[str, Any]:
        """Covert function with Basic metric data mode."""
        return {
            "run_result": self.data.run_result,
            "execution_time_ms": self.data.execution_time_ms,
            "engine": self.data.metric_engine.model_dump(by_alias=True),
            "source": self.data.metric_source.model_dump(by_alias=True),
            "transform": self.data.metric_transform.model_dump(by_alias=True),
            "sink": self.data.metric_sink.model_dump(by_alias=True),
            "custom_metric": self.custom_metric,
        }


class FullConvertor(BaseConvertor):

    def convert(self) -> dict[str, Any]:
        """Covert function with Full metric data mode."""
        return json.loads(self.data.model_dump_json(by_alias=True)) | {
            "custom_metric": self.custom_metric
        }


Convertor = Literal[
    "basic",
    "full",
]
ConvertType = TypeVar("ConvertType", bound=BaseConvertor)
CONVERTOR_REGISTRY: dict[Convertor, type[ConvertType]] = {
    "full": FullConvertor,
    "basic": BasicConvertor,
}



from abc import ABC, abstractmethod

from pyarrow import Schema, Table

from ....__types import DictData
from ....models import MetricOperatorOrder
from ...__abc import BaseTransform


class BaseArrowTransform(BaseTransform, ABC):

    @abstractmethod
    def apply(
        self,
        df: Table,
        engine: DictData,
        metric: MetricOperatorOrder,
        **kwargs,
    ) -> Table: ...

    @staticmethod
    def sync_schema(pre: Schema, post: Schema, metric, **kwargs) -> None: ...



"""Base Model modules keep all necessary models that use for each engine,
metric, and convertor.
"""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Literal, TypedDict, Union

from pydantic import BaseModel, ConfigDict, Field
from pydantic.functional_validators import model_validator
from typing_extensions import NotRequired, Self

from jett.__types import DictData
from jett.utils import get_dt_latest, get_dt_now


class Shape(BaseModel):
    """Shape model for keeping simple log for the Source and Sink metric model.

    Examples:
        Create Shape directly
        >>> Shape(rows=1, columns=2)

        Create Shape from_tuple classmethod.
        >>> Shape.from_tuple((1, 2))
    """

    rows: int = Field(default=0, ge=0, description="A row/record number.")
    columns: int = Field(default=0, ge=0, description="A column/field number.")

    @classmethod
    def from_tuple(cls, data: tuple[int, int]) -> Self:
        """Construct Shape model with a pair of rows and columns.

        Args:
            data: tuple[int, int]: A pair of rows and columns number.
        """
        rows, columns = data
        return cls(rows=rows, columns=columns)


class ColDetail(BaseModel):
    """A shared data structure of column detail that should be use only in
    Result model.

    Examples:
        >>> ColDetail(name="id", dtype="integer")
    """

    name: str = Field(description="A name of column.")
    dtype: str = Field(description="A data type of column in string format.")


class Result(BaseModel):
    """A shared data structure of result of Jett tool. This model will use to be
    an output from tool execution step.

    Examples:
        >>> Result(
        ...     data=[{"id": 1}, {"id": 2}],
        ...     columns=[ColDetail(name="id", dtype="integer")],
        ... )
    """

    data: list[Any] = Field(
        default_factory=list,
        description="A sample data that already save to the sink.",
    )
    columns: list[ColDetail] = Field(default_factory=list)
    schema_dict: dict[str, Any] = Field(default_factory=dict)


class EmitCond(str, Enum):
    """An Enum for Metric Emitter Condition."""

    ONLY_SUCCESS = "only_success"
    ONLY_FAILED = "only_failed"
    ALWAYS = "always"


ALWAYS = EmitCond.ALWAYS
ONLY_SUCCESS = EmitCond.ONLY_SUCCESS
ONLY_FAILED = EmitCond.ONLY_FAILED


Operation = Literal[">", ">=", "<", "<=", "=", "!="]


class BasicFilter(BaseModel):
    """A Configuration Model for Basic Filter."""

    col: str = Field(description="A column name.")
    op: Operation = Field(
        description=(
            "An operation type for filtering, an operator e.g. >, >=, <, <=, "
            "or =."
        )
    )
    value: str | None = Field(
        default=None,
        description="a value to be compared",
    )
    value_expression: str | None = Field(
        default=None,
        description="a value expression to be compared",
    )

    @model_validator(mode="after")
    def __validate_value_and_value_expression(self):
        """Validates value and value_expression should not be provided at the
        same time.
        """
        if (self.value is None) == (self.value_expression is None):
            raise ValueError(
                "Please choose either value or value_expression or specify one "
                "between value or value_expression."
            )
        return self

    def get_str_cond(self) -> str:
        """Get the string condition as a string from model."""
        return f"( {' '.join(self.get_cond())} )"

    def get_cond(self) -> tuple[str, str, str]:
        """Get order of condition."""
        return (
            (self.col, self.op, self.value_expression)
            if self.value_expression
            else (self.col, self.op, self.value)
        )


class BaseMetricData(BaseModel):
    """Base Metric class for any inheritance class of metric data.

    Methods:
        add: Add any custom data to the extra field.
    """

    extras: dict[str, Any] = Field(
        default_factory=dict,
        description="An extra metric data.",
    )
    start_dt: datetime = Field(
        default_factory=get_dt_now,
        description="A start datetime from start create this metric model.",
    )
    end_dt: datetime | None = Field(
        default=None,
        description="An end datetime from the handler method.",
    )
    latency_ms: float = Field(
        default=0,
        description="A latency from start to end.",
    )

    def add(self, key: str, value: Any) -> Self:
        """Add any custom data to the extras field by given a pair of key and
        value.

        Args:
            key (str): A key of this extras metric.
            value: A value of this extras metric.
        """
        self.__dict__["extras"][key] = value
        return self

    def finish(self) -> Self:
        """Set the end_dt field."""
        self.end_dt = get_dt_now()
        self.latency_ms = (self.end_dt - self.start_dt).total_seconds()
        return self


class MetricSource(BaseMetricData):
    """Metric Source model."""

    type: str | None = None
    read_row_count: int = 0
    read_column_count: int = 0
    inlet: tuple[str, str] | None = None


class MetricOperator(BaseMetricData):
    """Metric Operator model."""

    type: Literal["order"]
    trans_op: str


class MetricOperatorOrder(MetricOperator):
    """Metric Operator Transform order model."""

    transform_pre: dict[str, Any] = Field(default_factory=dict)
    transform_post: dict[str, Any] = Field(default_factory=dict)


class MetricOperatorGroup(BaseMetricData):
    """Metric Operator Transform group model."""

    type: Literal["group"]
    trans_op: Literal["group"] = "group"
    transforms: list[MetricOperator] = Field(default_factory=list)
    transform_pre: dict[str, Any] = Field(default_factory=dict)
    transform_post: dict[str, Any] = Field(default_factory=dict)


MetricOperatorTransform = Annotated[
    Union[
        MetricOperatorGroup,
        MetricOperatorOrder,
    ],
    Field(
        discriminator="type",
        description="A metric operator transform that dynamic with the `type`.",
    ),
]


class MetricTransform(BaseMetricData):
    transforms: list[MetricOperatorTransform] = Field(
        default_factory=list,
        description="List of transform operator metric.",
    )


class MetricSink(BaseMetricData):
    destination: str = ""
    write_row_count: int = 0
    write_column_count: int = 0
    outlet: tuple[str, str] | None = None
    type: str | None = None
    mode: str | None = Field(
        default=None,
        description="A sink mode that use like override or append",
    )


class MetricEngine(BaseMetricData):
    app_id: str | None = Field(
        default=None,
        description="An application ID that use on this engine session.",
    )
    type: str | None = Field(
        default=None,
        description="An engine type.",
    )


class MetricData(BaseModel):
    """A shared data structure of metric model."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    run_result: bool = Field(
        default=False,
        description=(
            "A running result success to catch from execution if it true."
        ),
    )
    execution_time_ms: float = Field(
        default=0.0,
        description=(
            "An execution time that start when calling the execute method."
        ),
    )
    execution_start_time: datetime = Field(default_factory=get_dt_now)
    execution_end_time: datetime = Field(default_factory=get_dt_latest)
    exception: Exception | BaseException | None = None
    exception_name: str | None = None
    exception_traceback: str | None = None

    metric_engine: MetricEngine = Field(
        default_factory=MetricEngine,
        description="A metric engine that was created on the `handle_execute` method.",
    )
    metric_source: MetricSource = Field(default_factory=MetricSource)
    metric_transform: MetricTransform = Field(
        default_factory=MetricTransform,
        description="A metric transform that was created after `handle_apply` method.",
    )
    metric_sink: MetricSink = Field(default_factory=MetricSink)


class Context(TypedDict, total=False):
    """Context dict type for pre-validate for Jett tool execution."""

    # NOTE: Before execute
    author: str | None
    owner: str | None
    parent_dir: Path

    # NOTE: After execute
    run_result: NotRequired[bool]
    execution_time_ms: NotRequired[float]
    execution_start_time: NotRequired[datetime]
    execution_end_time: NotRequired[datetime]
    exception: NotRequired[Exception | BaseException]
    exception_name: NotRequired[str]
    exception_traceback: NotRequired[str]
    post_execution_output: NotRequired[Any]

    # NOTE: Updated from that layer.
    metric_engine: NotRequired[MetricEngine]
    metric_source: NotRequired[MetricSource]
    metric_sink: NotRequired[MetricSink]

    # NOTE: Not necessary keys for transform metric.
    metric_group_transform: NotRequired[Any]
    metric_operator: NotRequired[Any]

    metric_transform: NotRequired[MetricTransform]
    metric_extras: NotRequired[DictData]



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field

from ....__types import DictData
from ....models import Context, MetricOperatorOrder, MetricOperatorTransform
from ...__abc import BaseTransform

if TYPE_CHECKING:
    from polars import DataFrame, DataType, Schema
    from polars import Expr as Column

    PairCol = tuple[Column, str]
    AnyApplyGroupOutput = PairCol | list[PairCol]
    AnyApplyOutput = AnyApplyGroupOutput | DataFrame

logger = logging.getLogger("jett")


def get_map_dtypes() -> dict[str, type[DataType]]:
    import polars as pl

    return {
        "string": pl.String,
        "boolean": pl.Boolean,
        # "integer": ...,
        # "double": ...,
        # "timestamp": ...,
    }


def is_pair_expr(value: PairCol | Any) -> bool:
    """Change value is a pair of Column and string.

    Args:
        value (PairCol | Any): A pair of Column and its name or any value.

    Returns:
        bool: True if an input value is a pair of Column and alias.
    """
    return (
        isinstance(value, tuple)
        and len(value) == 2
        and isinstance(value[0], Column)
        and isinstance(value[1], str)
    )


class BasePolarsTransform(BaseTransform, ABC):
    """Base Polars Transform abstract model"""

    @abstractmethod
    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        metric: MetricOperatorOrder,
        **kwargs,
    ) -> AnyApplyOutput:
        """Apply operator transform abstraction method.

        Args:
            df (DataFrame): A Polars DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricOperatorOrder): A metric transform that was set from
                handler step for passing custom metric data.

        Returns:
            AnyApplyOutput: That able to be:
                - PairCol
                - list[PairCol]: A list of pair Expr and alias name.
                - DataFrame: A Polars DataFrame.
        """

    def handle_apply(
        self,
        df: DataFrame,
        context: Context,
        engine: DictData,
        **kwargs,
    ) -> DataFrame:
        """Handle the Operator Apply result output from its custom apply that
        can make different type of result.

        Args:
            df (DataFrame): A Polars DataFrame.
            context: (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            **kwargs:

        Returns:
            DataFrame: A applied Polars DataFrame.
        """
        logger.info(f"👷🔧 Handle Apply Operator: {self.op!r}")
        metric = MetricOperatorOrder(type="order", trans_op=self.op)
        context["metric_operator"].append(metric)
        try:
            pre_schema: Schema = df.schema
            output: AnyApplyOutput = self.apply(
                df, engine=engine, metric=metric, **kwargs
            )
            if is_pair_expr(output):
                df = df.with_columns(output[0].alias(output[1]))
            elif isinstance(output, list) and all(
                is_pair_expr(i) for i in output
            ):
                df = df.with_columns(i[0].alias(i[1]) for i in output)
            elif isinstance(output, DataFrame):
                df: DataFrame = output
            else:
                metric.add(key=self.op, value=output)
                logger.info(f"Set metric from func {self.op} completely")

            # NOTE: Sync schema before return.
            self.sync_schema(pre_schema, df.schema, metric=metric)
            return df
        finally:
            self.post_apply(engine=engine, **kwargs)
            metric.finish()

    @staticmethod
    def sync_schema(
        pre: Schema, post: Schema, metric: MetricOperatorTransform, **kwargs
    ) -> None: ...


class ColMap(BaseModel):
    """Column Map model."""

    name: str = Field(description="A new column name.")
    source: str = Field(
        description="A source column statement before alias with alias.",
    )



from typing import Annotated, Union

from pydantic import Field

from .files.local import LocalCSVFile, LocalJsonFile

# LocalFile = Annotated[
#     Union[
#         LocalCSVFile,
#         LocalJsonFile,
#     ],
#     Field(discriminator="file_format"),
# ]

Source = Annotated[
    Union[
        LocalCSVFile,
        LocalJsonFile,
    ],
    Field(discriminator="file_format"),
]



from __future__ import annotations

import logging
import os
from abc import ABC
from typing import TYPE_CHECKING, Literal

from pydantic import Field, SecretStr
from pydantic.functional_validators import model_validator

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession

from .....__types import DictData
from .....models import BasicFilter, Shape
from .....utils import bool2str
from ....__abc import BaseSource
from ...utils import is_remote_session

logger = logging.getLogger("jett")


class BaseS3File(BaseSource, ABC):
    """AWS S3 CSV file data source."""

    type: Literal["s3", "vos"]
    path: str
    file_format: str
    access_key: SecretStr | None = None
    secret_key: SecretStr | None = None
    session_token: SecretStr | None = None
    s3_endpoint: str
    s3_region: str | None = None
    s3_path_style: bool | None = True
    impl: str = "org.apache.hadoop.fs.s3a.S3AFileSystem"
    s3_credentials_provider: str = (
        "org.apache.hadoop.fs.s3a.SimpleAWSCredentialsProvider"
    )
    filter: list[BasicFilter] = Field(default_factory=list)

    @model_validator(mode="after")
    def __check_credential(self):
        """validate credential."""
        _access_key = os.getenv("AWS_ACCESS_KEY_ID", self.access_key)
        _secret_key = os.getenv("AWS_SECRET_ACCESS_KEY", self.secret_key)
        _session_token = os.getenv("AWS_SESSION_TOKEN", self.session_token)

        if self.s3_credentials_provider in (
            "org.apache.hadoop.fs.s3a.SimpleAWSCredentialsProvider",
            "org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider",
        ):
            if _access_key is None or _secret_key is None:
                raise ValueError(
                    "please supply access_key (env: AWS_ACCESS_KEY_ID) and "
                    "secret_key (env: AWS_SECRET_ACCESS_KEY)"
                )

            if (
                self.s3_credentials_provider
                == "org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider"
            ):
                if _session_token is None:
                    raise ValueError(
                        "please supply session_token (AWS_SESSION_TOKEN)"
                    )

        return self

    def set_s3_connection(self, spark: SparkSession) -> None:
        """Set s3 connection in SparkContext.

        Args:
            spark (SparkSession): The current spark session object that will use
                to create a DataFrame Reader object.
        """
        logger.info("set fs.s3a.impl=%s", self.impl)
        logger.info("set fs.s3a.endpoint=%s", self.s3_endpoint)
        logger.info("set fs.s3a.path.style.access=true")
        logger.info(
            "set fs.s3a.aws.credentials.provider=%s",
            self.s3_credentials_provider,
        )

        hadoop_conf = spark.sparkContext._jsc.hadoopConfiguration()
        hadoop_conf.set("fs.s3a.impl", self.impl)
        hadoop_conf.set("fs.s3a.endpoint", self.s3_endpoint)
        if self.s3_path_style:
            hadoop_conf.set("fs.s3a.path.style.access", "true")
        hadoop_conf.set(
            "fs.s3a.aws.credentials.provider",
            self.s3_credentials_provider,
        )

        if self.access_key is not None:
            hadoop_conf.set(
                "fs.s3a.access.key",
                self.access_key.get_secret_value(),
            )
            hadoop_conf.set(
                "fs.s3a.secret.key",
                self.secret_key.get_secret_value(),
            )

        if self.session_token is not None:
            hadoop_conf.set(
                "fs.s3a.session.token",
                self.session_token.get_secret_value(),
            )

        if self.s3_region is not None:
            hadoop_conf.set("fs.s3a.endpoint.region", self.s3_region)

    def inlet(self) -> tuple[str, str]:
        return "s3", self.path


class S3CSVFile(BaseS3File):
    file_format: Literal["csv"]
    delimiter: str = "|"
    header: bool = Field(default=True)

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        """Apply S3 CSV Loading process with the Spark session object."""
        spark: SparkSession = engine["spark"]
        if is_remote_session(spark):
            raise RuntimeError(
                f"Source: {self.type} does not support Spark Connect Session"
            )
        logger.info(f"🚰 Source - Start Load S3: {self.path}")
        self.set_s3_connection(spark=spark)
        reader = spark.read.format(self.file_format)

        df: DataFrame = (
            reader.option("header", bool2str(self.header))
            .option("delimiter", self.delimiter)
            .option("inferSchema", "true")
            .load(self.path)
        )

        if self.filter:
            df = df.filter(" and ".join(f.get_str_cond() for f in self.filter))

        if self.sample_records:
            logger.info(f"🚰 Source - apply limit: {self.sample_records}")
            df = df.limit(self.sample_records)

        rows: int = df.count()
        shape: tuple[int, int] = (rows, len(df.columns))
        return df, Shape.from_tuple(shape)


class S3JsonFile(BaseS3File):
    file_format: Literal["json"]
    multiline: bool = False

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        spark: SparkSession = engine["spark"]
        if is_remote_session(spark):
            raise RuntimeError(
                f"Source: {self.type} does not support Spark Connect Session"
            )
        logger.info(f"🚰 Source - Start Load S3: {self.path}")
        self.set_s3_connection(spark=spark)
        reader = spark.read.format(self.file_format)

        df: DataFrame = (
            reader.option("inferSchema", "true")
            .option("multiline", bool2str(self.multiline))
            .load(self.path)
        )

        if self.filter:
            df = df.filter(" and ".join(f.get_str_cond() for f in self.filter))

        if self.sample_records:
            logger.info(f"🚰 Source - apply limit: {self.sample_records}")
            df = df.limit(self.sample_records)

        rows: int = df.count()
        shape: tuple[int, int] = (rows, len(df.columns))
        return df, Shape.from_tuple(shape)



from __future__ import annotations

from typing import Annotated, Union

from pydantic import Field

from .arrow import Arrow
from .daft import Daft
from .dbt import Dbt
from .duckdb import DuckDB
from .polars import Polars
from .spark import Spark

Engine = Annotated[
    Union[
        Arrow,
        Daft,
        Polars,
        DuckDB,
        Spark,
        Dbt,
    ],
    Field(
        discriminator="type",
        description="The main engine registry models for operator model.",
    ),
]



from typing import Annotated, Union

from pydantic import Field

from .files.local import (
    LocalCsvFile,
    LocalJsonFile,
)

Source = Annotated[
    Union[
        LocalCsvFile,
        LocalJsonFile,
    ],
    Field(discriminator="file_format"),
]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.types import StructField, StructType

from .....utils import clean_string
from ...schema_change import (
    clean_col_except_item_from_extract_array,
    evaluate_schema_change,
    summarize_changes,
)
from ...utils import (
    column_has_only_null_or_empty_array,
    remove_column_from_struct,
    replace_column_dtype_in_nested_schema,
)
from .utils import (
    clean_hidden_partition,
    extract_hidden_partition,
    gen_alter_cols,
    gen_alter_partitions,
    get_table_partition_columns,
    is_column_safe_to_alter,
)

logger = logging.getLogger("jett")


class EvolverModel(BaseModel):
    """Iceberg Table Evolver, A Class for Evolution of Iceberg Table Schema and
    Partition.
    """

    database: str = Field(description="A database name.")
    table_name: str = Field(description="A table name.")
    schema_evolve_behavior: Literal["trust_source", "append", "trust_sink"] = (
        Field(
            description="behavior of how to evolve schema, please see IcebergModel"
        )
    )
    allow_evolve_with_missing_columns: bool = Field(
        default=False,
        description=(
            "a boolean flag, if True, the workflow will not reject when they "
            "found missing column and continue to evolve the schema"
        ),
    )
    allow_cast_safe_dtype: bool = Field(
        default=False,
        description=(
            "a boolean flag, if True, it allows to change data type of "
            "compatible columns (e.g. int (column in df) can be converted "
            "into double (column in table))"
        ),
    )
    partition_by: list[str] = Field(default_factory=list)


class Evolver:

    def __init__(
        self,
        model: EvolverModel,
        *,
        df: DataFrame,
        spark: SparkSession,
    ) -> None:
        self.model: EvolverModel = model
        self.df: DataFrame = df
        self.spark: SparkSession = spark

        # NOTE: Set internal arguments
        self.is_schema_diff: bool = False
        self.is_table_schema_changed: bool = False
        self.is_partition_diff: bool = False
        self.is_table_partition_changed: bool = False
        self.current_partition_spec: list[str] = []
        self.transform_dict_safe_cast: dict[str, StructField] = {}
        self.detailed_changes: list[dict] = []
        self.list_sql_alter: list[str] = []

        self.transform_schema_to_drop: list[str] = []
        self.transform_schema_after_drop: StructType | None = None

    def __execute_add_column(self, to_add_col_cmds):
        logger.info("total new columns: %s", len(to_add_col_cmds))
        to_add_cols_cmd_str = ",\n".join(to_add_col_cmds)
        add_new_cols_stmt = f"""
        ALTER TABLE {self.model.database}.{self.model.table_name}
        ADD COLUMNS (
        {to_add_cols_cmd_str}
        )
        """
        add_new_cols_stmt = clean_string(text=add_new_cols_stmt)
        logger.info(add_new_cols_stmt)
        self.list_sql_alter.append(add_new_cols_stmt)
        self.spark.sql(add_new_cols_stmt)

    def __execute_alter_column(self, to_alter_col_cmds):
        # see column types that Iceberg can covert type safely
        # https://iceberg.apache.org/docs/latest/spark-ddl/#alter-table-alter-column
        logger.info(
            "total columns to be changed data type: %s", len(to_alter_col_cmds)
        )
        for cmd in to_alter_col_cmds:
            alter_col_stmt = (
                f"ALTER TABLE {self.model.database}.{self.model.table_name} "
                f"ALTER COLUMN {cmd}"
            )
            logger.info(alter_col_stmt)
            self.list_sql_alter.append(alter_col_stmt)
            self.spark.sql(alter_col_stmt)

    def __execute_drop_column(self, to_drop_col_cmds):
        logger.info("total columns to be dropped: %s", len(to_drop_col_cmds))
        for cmd in to_drop_col_cmds:
            drop_col_stmt = (
                f"ALTER TABLE {self.model.database}.{self.model.table_name} "
                f"DROP COLUMN {cmd}"
            )
            logger.info(drop_col_stmt)
            self.list_sql_alter.append(drop_col_stmt)
            self.spark.sql(drop_col_stmt)

    def _evolve_schema_trust_source(
        self,
        to_add_col_cmds: list[str],
        to_alter_col_cmds: list[str],
        to_drop_col_cmds: list[str],
    ) -> None:
        """
        Alter table columns, based on new schema from source DataFrame
        The function supports 3 operations, which are
        1. alter type of columns
        2. add new columns
        3. and drop columns
        """

        if len(to_alter_col_cmds) > 0:
            self.__execute_alter_column(to_alter_col_cmds)

        if len(to_add_col_cmds) > 0:
            self.__execute_add_column(to_add_col_cmds)

        if len(to_drop_col_cmds) > 0:
            self.__execute_drop_column(to_drop_col_cmds)

        if (
            len(to_alter_col_cmds) > 0
            or len(to_add_col_cmds) > 0
            or len(to_drop_col_cmds) > 0
        ):
            self._is_table_schema_changed = True

        logger.info("schema change completed")

    def _evolve_schema_append(
        self,
        to_add_col_cmds: list[str],
        to_alter_col_cmds: list[str],
        to_drop_col_cmds: list[str],
    ) -> None:
        """
        Alter table columns, based on new schema from source Dataframe
        The function supports only 2 operations, which are
        1. alter type of columns
        2. add new columns
        Because of append mode doesn't allow to remove or drop any column or field
        """

        if self.model.allow_evolve_with_missing_columns:
            if len(to_alter_col_cmds) > 0:
                self.__execute_alter_column(to_alter_col_cmds)

            if len(to_add_col_cmds) > 0:
                self.__execute_add_column(to_add_col_cmds)
        else:
            if len(to_drop_col_cmds) > 0:
                missing_cols = "\n".join(to_drop_col_cmds)
                raise RuntimeError(
                    f"please add the missing columns: {missing_cols}"
                )

            if len(to_alter_col_cmds) > 0:
                self.__execute_alter_column(to_alter_col_cmds)

            if len(to_add_col_cmds) > 0:
                self.__execute_add_column(to_add_col_cmds)

        if len(to_alter_col_cmds) > 0 or len(to_add_col_cmds) > 0:
            self._is_table_schema_changed = True

        logger.info("schema change completed")

    def _set_transform_details_for_safe_column_change_dtype(
        self,
        current_dtype: StructField,
        to_be_dtype: StructField,
        parent_cols: list[str],
    ) -> None:
        """Set the transform dict for columns that are safe to be change data
        type.
        """
        from pyspark.sql.types import StructField, StructType

        if len(parent_cols) > 0:
            root_column = parent_cols[0]
            root_schema = self.df.select(root_column).schema[0]
            if root_column in self.transform_dict_safe_cast.keys():
                root_schema = self.transform_dict_safe_cast[root_column]

            target_column = None
            to_dtype = current_dtype
            if isinstance(to_be_dtype, StructField):
                target_column = to_be_dtype.name
                to_dtype = current_dtype.dataType

            replaced_schema = replace_column_dtype_in_nested_schema(
                schema=root_schema,
                parent_cols=parent_cols.copy(),
                target_column=target_column,
                to_dtype=to_dtype,
            )
        else:
            root_column = current_dtype.name
            replaced_schema = current_dtype

        self.transform_dict_safe_cast[root_column] = replaced_schema

        # set schema for drop columns, used in method _is_column_contain_only_null_or_empty_array
        if self.transform_schema_after_drop is None:
            self.transform_schema_after_drop = self.df.schema

        json_schema = self.transform_schema_after_drop.jsonValue()
        to_fixed_index = [
            i
            for i, field in enumerate(json_schema["fields"])
            if field["name"] == root_column
        ][0]
        json_schema["fields"][to_fixed_index] = replaced_schema.jsonValue()
        self.transform_schema_after_drop = StructType.fromJson(json_schema)

    def _is_column_safe_to_be_changed_dtype(
        self,
        current_dtype,
        to_be_dtype,
        parent_cols: list[str],
    ) -> bool:
        """
        check compatibility of changing data type for common cases
        support cases are
        - cast integer to double
        - cast integer to float
        - cast integer to bigint
        note: current_dtype is the current column's data type in the table
              to_be_dtype is the column's data type in the dataframe
        """
        from pyspark.sql.types import (
            DoubleType,
            FloatType,
            IntegerType,
            LongType,
            StructField,
        )

        is_safe_update = False
        _cur_dtype = current_dtype
        _tobe_dtype = to_be_dtype
        if isinstance(current_dtype, StructField):
            _cur_dtype = current_dtype.dataType
            _tobe_dtype = to_be_dtype.dataType

        if isinstance(_tobe_dtype, IntegerType):
            if isinstance(_cur_dtype, DoubleType):
                is_safe_update = True
            elif isinstance(_cur_dtype, FloatType):
                is_safe_update = True
            elif isinstance(_cur_dtype, LongType):
                is_safe_update = True

            # _cur_dtype is DecimalType, this will not support due to not know the precision and scale
            # it can cause null value after cast if scale is small

        if is_safe_update:
            self._set_transform_details_for_safe_column_change_dtype(
                current_dtype=current_dtype,
                to_be_dtype=to_be_dtype,
                parent_cols=parent_cols,
            )

        return is_safe_update

    def _is_column_contain_only_null_or_empty_array(
        self,
        to_be_dtype,
        parent_cols: list[str],
    ) -> bool:
        """
        check that the given column contain only null value or empty array or not
        if it's empty array, append the column to the list and wait for drop column operation
        """
        from pyspark.sql.types import ArrayType, StructField, StructType

        is_safe_update = False

        # allow to check only primitive type
        if isinstance(to_be_dtype, (StructType, ArrayType)):
            return is_safe_update

        columns = parent_cols.copy()
        if isinstance(to_be_dtype, StructField):
            columns.append(to_be_dtype.name)
            columns_str = ".".join(columns)
        else:
            # primitive type, remove last column which is element
            columns_str = ".".join(columns[:-1])

        column_without_element = clean_col_except_item_from_extract_array(
            columns=columns
        )
        is_safe_update = column_has_only_null_or_empty_array(
            df=self.df, column=".".join(column_without_element)
        )
        if is_safe_update:
            schema = self.transform_schema_after_drop
            if schema is None:
                schema = self.df.schema

            self.transform_schema_after_drop = remove_column_from_struct(
                schema=schema, column=columns_str
            )
            self.transform_schema_to_drop.append(columns_str)

        return is_safe_update

    def validate_column_changes_compatibility(self) -> None:
        """validate the compatibility of columns changes before alter operation."""
        disallow_changes_unsafe_alter = []
        for s in self.detailed_changes:
            change_type = s.get("type")
            target_struct_type = s.get("target_struct_type", None)
            source_struct_type = s.get("source_struct_type", None)
            parent_cols = s.get("parent_cols", [])

            if change_type == "alter_col":
                is_safe_update = is_column_safe_to_alter(
                    current_dtype=target_struct_type,
                    to_be_dtype=source_struct_type,
                )
                if is_safe_update is False and self.model.allow_cast_safe_dtype:
                    is_safe_update = self._is_column_safe_to_be_changed_dtype(
                        current_dtype=target_struct_type,
                        to_be_dtype=source_struct_type,
                        parent_cols=parent_cols,
                    )
                if (
                    is_safe_update is False
                    and self.model.allow_evolve_with_missing_columns
                ):
                    is_safe_update = (
                        self._is_column_contain_only_null_or_empty_array(
                            to_be_dtype=source_struct_type,
                            parent_cols=parent_cols,
                        )
                    )

                if not is_safe_update:
                    disallow_changes_unsafe_alter.append(s)

        if len(disallow_changes_unsafe_alter) > 0:
            _errors = []
            for s in disallow_changes_unsafe_alter:
                parent_col = ".".join(s["parent_cols"])
                source_str = s["source_struct_type"].simpleString()
                target_str = s["target_struct_type"].simpleString()
                _error = f"from {target_str} to {source_str}"
                if parent_col != "":
                    _error = f"{_error} (parent column={parent_col})"
                _errors.append(_error)
            err = "\n".join(_errors)
            raise RuntimeError(
                f"found unsafe columns to be altered data type\n{err}"
            )

    def fetch_detailed_changes(self) -> list[dict]:
        """Wrapper of evaluate_schema_change function that use for getting
        detailed changes of each different column or field from DataFrame schema
        compare to table schema.
        """
        source_schema: StructType = self.df.schema
        table_schema: StructType = self.spark.sql(
            f"SELECT * FROM {self.model.database}.{self.model.table_name} "
            f"LIMIT 0"
        ).schema
        return evaluate_schema_change(
            src_schema=source_schema,
            tgt_schema=table_schema,
        )

    def evolve_schema(self) -> None:
        """Evolve table schema based on incoming dataframe and evolve behavior."""
        from pyspark.sql.functions import col

        logger.info(
            f"Schema evolve behavior: {self.model.schema_evolve_behavior}"
        )
        table_schema = self.spark.sql(
            f"SELECT * FROM {self.model.database}.{self.model.table_name}"
        ).schema
        schema_diff: bool = self.df.schema != table_schema
        logger.info(f"schema diff: {schema_diff}")
        if schema_diff:
            self.is_schema_diff = True
            self.detailed_changes = self.fetch_detailed_changes()
            if len(self.detailed_changes) == 0:
                logger.info(
                    "schema does not change, only nullable or metadata of "
                    "schema is difference, skip evolution"
                )
                return

            logger.info("Validate changes of columns")
            self.validate_column_changes_compatibility()

            # cast data type for compatible type change
            if len(self.transform_dict_safe_cast) > 0:
                logger.info(
                    "found %s columns that are safe to be changed their own data type",
                    len(self.transform_dict_safe_cast),
                )
                logger.info("apply cast data type for safe columns")
                _transform_dict = {}
                for k, v in self.transform_dict_safe_cast.items():
                    _transform_dict[k] = col(k).cast(v.dataType)
                    logger.info(
                        "target col: %s, cast as: %s",
                        k,
                        v.dataType.simpleString(),
                    )
                self.df = self.df.withColumns(_transform_dict)

            if len(self.transform_schema_to_drop) > 0:
                logger.info(
                    "found %s columns contain only null or empty array",
                    len(self.transform_schema_to_drop),
                )
                logger.info(
                    "affected columns (dropped from df): %s",
                    ", ".join(self.transform_schema_to_drop),
                )
                self.df = self.df.to(self.transform_schema_after_drop)

            # NOTE: Refresh the changes
            self.detailed_changes = self.fetch_detailed_changes()
            match self.model.schema_evolve_behavior:
                case "trust_source":
                    logger.info("evolve schema based on source dataframe")
                    self._evolve_schema_trust_source(
                        *gen_alter_cols(changes=self.detailed_changes)
                    )
                case "append":
                    logger.info(
                        "append schema only (add new columns and alter columns)"
                    )
                    self._evolve_schema_append(
                        *gen_alter_cols(changes=self.detailed_changes)
                    )

    def apply_alter(self, proc: list[str]):
        for p in proc:
            logger.info(f"Apply: {p}")
            self.list_sql_alter.append(p)
            self.spark.sql(p)

    def evolve_partition(self) -> None:
        """Evolve partition."""
        self.current_partition_spec = get_table_partition_columns(
            spark=self.spark,
            db_name=self.model.database,
            table_name=self.model.table_name,
        )
        _source = [
            clean_hidden_partition(partition_spec=p)
            for p in self.model.partition_by
        ]
        _table = [
            clean_hidden_partition(partition_spec=p)
            for p in self.current_partition_spec
        ]
        partition_diff: bool = sorted(_source) != sorted(_table)
        logger.info(f"partition diff: {partition_diff}")
        if not partition_diff:
            return

        self.is_partition_diff = True
        logger.info("evolve partition")
        logger.info(
            f"Current table partition spec: {self.current_partition_spec}"
        )
        logger.info(f"new partition specification: {self.model.partition_by}")
        to_add, to_replace, to_drop = gen_alter_partitions(
            database=self.model.database,
            table_name=self.model.table_name,
            source_partition=extract_hidden_partition(self.model.partition_by),
            table_partition=extract_hidden_partition(
                self.current_partition_spec
            ),
        )

        if to_drop:
            logger.info(f"total partitions to be dropped: {len(to_drop)}")
            self.apply_alter(to_drop)

        if to_replace:
            logger.info(
                f"total existing partitions to be replaced: {len(to_replace)}"
            )
            self.apply_alter(to_replace)

        if to_add:
            logger.info(f"total new partitions: {len(to_add)}")
            self.apply_alter(to_add)

        self.is_table_partition_changed = True
        logger.info("partition is evolved successfully")

    def get_simple_detail_changes(
        self,
    ) -> tuple[list[dict], list[dict], list[dict]]:
        """
        get summarize of column changes
        """
        alter_list, add_list, drop_list = summarize_changes(
            changes=self.detailed_changes
        )
        if self.model.allow_evolve_with_missing_columns:
            drop_list = []
        if self.model.allow_cast_safe_dtype:
            alter_list = []
        return alter_list, add_list, drop_list



from __future__ import annotations

import base64
from typing import TYPE_CHECKING, Literal

from pydantic import SecretStr

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrame, SparkSession

    PairCol = tuple[Column, str]

from ....__types import DictData
from .__abc import BaseSparkTransform


class GCMDecrypt(BaseSparkTransform):
    """GCM Decrypt Transform model."""

    op: Literal["gcm_decrypt"]
    name: str
    source: str
    secret_key: SecretStr
    encode_mode: Literal["base64"] = "base64"
    block_size: int = 16

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> PairCol:
        """Decryption with GCM mode.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.
        """
        from Crypto.Cipher import AES
        from Crypto.Hash import HMAC, SHA256
        from pyspark.errors import PythonException
        from pyspark.sql.functions import col, udf
        from pyspark.sql.types import StringType

        def __decrypt_gcm(
            cipher_test: str,
            secret_key: str,
            bs: int,
            encode: Literal["base64"],
        ) -> str:  # pragma: no cover
            """Create Wrapped function that will use for making PySpark UDF
            function via Python API.

            Warning: Fixing the case of SPARK-5063.

            Returns: str
            """
            secret_key = secret_key.encode()
            nonce = HMAC.new(secret_key, msg=b"", digestmod=SHA256).digest()
            cipher = AES.new(secret_key, AES.MODE_GCM, nonce, mac_len=bs)

            if encode == "base64":
                decode_func = base64.b64decode
            else:
                raise PythonException(f"not support encode mode: {encode}")

            try:
                decoded_data = decode_func(cipher_test)
                base64_text = decoded_data[:-bs]
                mac_tag = decoded_data[-bs:]
                decrypted_text = cipher.decrypt_and_verify(
                    base64_text, mac_tag
                ).decode("utf8")

            except Exception as e:
                raise PythonException(
                    f"decryption error, encrypted value: {cipher_test}, "
                    f"error message: {e}"
                ) from e

            return decrypted_text

        secret: str = self.secret_key.get_secret_value()
        block_size: int = self.block_size
        encode_mode: Literal["base64"] = self.encode_mode
        decrypt_udf = udf(
            lambda text: __decrypt_gcm(
                cipher_test=text,
                secret_key=secret,
                bs=block_size,
                encode=encode_mode,
            ),
            StringType(),
        )
        return decrypt_udf(col(self.source)), self.name



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal

from pydantic.functional_validators import field_validator, model_validator

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

from ....__types import DictData
from ....models import BasicFilter, Shape
from ...__abc import BaseSource
from ..sql_parser import extract_table_names_from_query

logger = logging.getLogger("jett")


class Hive(BaseSource):
    """Hive Spark Source model."""

    type: Literal["hive", "iceberg"]
    database: str | None = None
    table_name: str | None = None
    query: str | None = None
    filter: list[BasicFilter] | None = None

    @field_validator("database", "table_name")
    def __validate_space(cls, value: Any) -> Any:
        if value is not None and isinstance(value, str) and " " in value:
            raise ValueError("cannot contain whitespace")
        return value

    @model_validator(mode="after")
    def __post_validate(self):
        if self.query:
            if self.database or self.table_name:
                raise ValueError(
                    "Cannot provide `query` and `db_name`/`table_name` at the "
                    "same time."
                )
        elif not self.database or not self.table_name:
            raise ValueError(
                "If `query` is not provided, both `db_name` and `table_name` "
                "must be specified."
            )

        return self

    def load(self, engine: DictData, **kwargs) -> tuple[Any, Shape]:
        spark: SparkSession = engine["spark"]
        if self.query:
            logger.info(f"loading source from: {self.query}")
            df = spark.sql(self.query)
        else:
            logger.info(
                f"loading source from: {self.database}.{self.table_name}",
            )
            spark.sql(f"use {self.database}")
            df = spark.table(self.table_name)

        if self.filter:
            df = df.filter(" and ".join(f.get_str_cond() for f in self.filter))

        if self.sample_records:
            logger.info(f"Source - apply limit: {self.sample_records}")
            df = df.limit(self.sample_records)

        rows: int = df.count()
        shape: tuple[int, int] = (rows, len(df.columns))
        return df, Shape.from_tuple(shape)

    def inlet(self) -> tuple[str, str]:
        if self.query:
            inlet_result = extract_table_names_from_query(self.query)
        else:
            inlet_result = f"{self.database}.{self.table_name}"
        return self.type, inlet_result



from pathlib import Path
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from duckdb import DuckDBPyRelation

from pydantic import Field

from .....__types import DictData
from .....models import MetricSource, Shape
from ....__abc import BaseSource


class LocalCsvFile(BaseSource):
    """Local CSV file data source."""

    type: Literal["local"]
    file_format: Literal["csv"]
    path: str
    delimiter: str = "|"
    header: bool = Field(default=True)
    sample_records: int | None = 200

    def load(
        self, engine: DictData, metric: MetricSource, **kwargs
    ) -> tuple["DuckDBPyRelation", Shape]:
        """Load CSV file to DuckDB Relation object."""
        import duckdb

        file_format: str = Path(self.path).suffix
        if file_format not in (".csv",):
            raise NotImplementedError(
                f"Local file format: {file_format!r} does not support."
            )
        df = duckdb.read_csv(
            path_or_buffer=self.path,
            delimiter=self.delimiter,
            header=self.header,
            sample_size=self.sample_records,
        )
        return df, Shape.from_tuple(df.shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path


class LocalJsonFile(BaseSource):
    """Local JSON file data source."""

    type: Literal["local"]
    file_format: Literal["json"]
    path: str = Field(description="A path of local JSON file.")
    format: Literal["newline_delimited", "array"] = "newline_delimited"

    def load(
        self, engine: DictData, metric: MetricSource, **kwargs
    ) -> tuple["DuckDBPyRelation", Shape]:
        import duckdb

        file_format: str = Path(self.path).suffix
        if file_format not in (".json",):
            raise NotImplementedError(
                f"Local file format: {file_format!r} does not support."
            )
        # ARCHIVE:
        # df = duckdb.sql(
        #     f"""SELECT * FROM read_json_objects(
        #         '{self.path}'
        #         , format = '{self.format}'
        #     ) AS original_data
        #     """
        # )
        df = duckdb.read_json(path_or_buffer=self.path, format=self.format)
        return df, Shape.from_tuple(df.shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal

from ....__types import DictData
from ....models import Shape
from ...__abc import BaseSink

if TYPE_CHECKING:
    from pyspark.sql import DataFrame

logger = logging.getLogger("jett")


class Empty(BaseSink):
    """Console Spark Sink model."""

    type: Literal["empty"]

    def save(
        self,
        df: DataFrame,
        engine: DictData,
        **kwargs,
    ) -> Any:
        """Save the result data to the Console."""
        logger.info("🎯 Sink - Start sync with empty")
        shape: tuple[int, int] = (df.count(), len(df.columns))
        return df, Shape.from_tuple(shape)

    def outlet(self) -> tuple[str, str]:
        return self.type, self.dest()

    def dest(self) -> str:
        return "empty"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from daft import DataFrame

from pydantic import Field
from pydantic.functional_validators import field_validator

from ... import Result
from ...__types import DictData
from ...models import ColDetail, Context, MetricEngine, MetricTransform
from ..__abc import BaseEngine
from .sink import Sink
from .source import Source
from .transform import Transform

logger = logging.getLogger("jett")


class Daft(BaseEngine):
    """Daft Engine Model."""

    type: Literal["daft"]
    sink: list[Sink] = Field(description="A list of Sink model.")
    source: Source
    transforms: list[Transform] = Field(default_factory=list)

    @field_validator(
        "sink",
        mode="before",
        json_schema_input_type=Sink | list[Sink],
    )
    def __prepare_sink(cls, data: Any) -> Any:
        """Prepare the sink field value that should be list of Sink model."""
        return [data] if not isinstance(data, list) else data

    def execute(
        self,
        context: Context,
        engine: DictData,
        metric: MetricEngine,
    ) -> DataFrame:
        logger.info("Start execute with Arrow engine.")
        df: DataFrame = self.source.handle_load(context, engine=engine)
        df: DataFrame = self.handle_apply(df, context, engine=engine)
        for sk in self.sink:
            sk.handle_save(df, context, engine=engine)
        return df

    def set_engine_context(self, context: Context, **kwargs) -> DictData:
        """Set Daft engine context."""
        return {"engine": self}

    def set_result(self, df: DataFrame, context: Context) -> Result:
        return Result(
            data=[],
            columns=[
                ColDetail(name=f.name, dtype=str(f.dtype)) for f in df.schema()
            ],
            schema_dict={f.name: f.dtype for f in df.schema()},
        )

    def apply(
        self,
        df: DataFrame,
        context: Context,
        engine: DictData,
        metric: MetricTransform,
        **kwargs,
    ) -> DataFrame:
        for op in self.transforms:
            df: DataFrame = op.handle_apply(df, context, engine=engine)
        return df



import json
import logging
from collections.abc import Sequence
from typing import Any

from airflow.models.baseoperator import BaseOperator
from airflow.utils.context import Context
from yaml import safe_load

from jett.tools import Tool

logger = logging.getLogger("jett")


class JettOperator(BaseOperator):
    """Jett Airflow Operator object."""

    template_fields: Sequence[str] = ("tool",)
    template_ext: Sequence[str] = (".tool",)

    ui_color: str = "#eb52b3"

    def __init__(self, tool: str, **kwargs) -> None:
        """Main initialize method."""
        super().__init__(**kwargs)
        self.tool: str = tool

    def execute(self, context: Context) -> Any:
        """Execute Jett tool."""
        logger.info("Debug context:")
        logger.info(f"{json.dumps(dict(context), default=str, indent=2)}")
        logger.info(f"Start run jett: {self.tool}")
        tool_data: dict[str, Any] = safe_load(self.tool)
        tool = Tool(config=tool_data)
        tool.execute()



from __future__ import annotations

import logging
from datetime import date
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrame, Row, SparkSession

    from ..sink import Sink

    PairCol = tuple[Column, str]

from pydantic import Field, PrivateAttr
from pydantic.functional_validators import field_validator, model_validator
from typing_extensions import Self

from ....__types import DictData
from ....utils import get_dt_now, get_random_str_unique, to_snake_case
from ..utils import (
    extract_col_with_pattern,
    extract_cols_without_array,
    is_remote_session,
    is_table_exist,
    replace_all_occurrences,
)
from .__abc import BaseSparkTransform
from .__models import RenameColMap

logger = logging.getLogger("jett")


class Expr(BaseSparkTransform):
    """Expression Operator Transform model."""

    op: Literal["expr"]
    name: str = Field(
        description=(
            "An alias name of this output of the query expression result store."
        )
    )
    query: str = Field(description="An expression query.")

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> PairCol:
        """Apply priority transform to expression the query.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.
        """
        from pyspark.sql.functions import expr

        return expr(self.query), self.name


class SQL(BaseSparkTransform):
    """SQL Operator Transform model."""

    op: Literal["sql_transform"]
    query: str = Field(
        description=(
            "A query statement that will use on the `sql` feature. If you want "
            "to interact with the current DataFrame, you should use it with "
            "`temp_table` table on this statement."
        )
    )

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply SQL.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            DataFrame:
        """
        temp_table_name: str = f"temp_table_{get_random_str_unique()}"
        query = replace_all_occurrences(
            self.query, "temp_table", temp_table_name
        )
        df.createOrReplaceTempView(temp_table_name)
        return spark.sql(query)


class RenameSnakeCase(BaseSparkTransform):
    op: Literal["rename_snakecase"]
    allow_group_transform: bool = False

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply to Rename all columns (support nested columns) to snake case format.
        by default the schema and data when reading from source is already ordered.

        Warnings:

            this transform can result in wrong data when the order of nested
        column is not an alphabetical order. the order of 1st lv column can be
        unordered.

        Notes:
            the columns can be ordered by using the logic.
            >>> df.select(sorted(df.columns))

        Examples:
            Case of unordered nested columns
            >>> from pyspark.sql.types import StructType, StructField, BooleanType
            >>> input_schema = StructType([
            ...     StructField("parent_struct", StructType([
            ...         StructField("FIRST_KEY", BooleanType(), True),
            ...         StructField("SECOND_KEY", BooleanType(), True),
            ...         ]),
            ...         True
            ...     )
            ... ])
            >>> input_data = [{
            ...     "parent_struct": {
            ...         "SECOND_KEY": False,
            ...         "FIRST_KEY": True
            ...     }
            ... }]

            Output after renaming columns will be:
            >>> new_schema = StructType([
            ...     StructField("parent_struct", StructType([
            ...         StructField("first_key", BooleanType(), True),
            ...         StructField("second_key", BooleanType(), True),
            ...         ]),
            ...         True
            ...     )
            ... ])
            >>> new_data = [{
            ...     "parent_struct": {
            ...         "first_key": False,
            ...         "second_key": True
            ...     }
            ... }]
        """
        from pyspark.sql.functions import col

        non_snake_case_cols = extract_col_with_pattern(
            schema=df.schema, patterns=["non_snake_case"]
        )
        if len(non_snake_case_cols) == 0:
            return df

        final_cols = []
        transform_dict = {}
        df = df.select(sorted(df.columns))  # always sort columns
        for schema in df.schema:
            new_col_name = None
            temp_dict = {}

            if schema.name.islower() is False or " " in schema.name:
                new_col_name = to_snake_case(schema.name)
                temp_dict[new_col_name] = col(schema.name)

            dtype = schema.dataType.simpleString()
            if dtype.islower() is False or " " in dtype:
                new_dtype = to_snake_case(dtype)
                if len(temp_dict) > 0:
                    val = temp_dict[new_col_name]
                    temp_dict[new_col_name] = val.cast(new_dtype)
                else:
                    temp_dict[schema.name] = col(schema.name).cast(new_dtype)

            if len(temp_dict) > 0:
                transform_dict = {**transform_dict, **temp_dict}
            else:
                final_cols.append(schema.name)

        if len(transform_dict) > 0:
            non_snake_case_cols = extract_col_with_pattern(
                schema=df.schema, patterns=["non_snake_case"]
            )
            logger.info(
                "found non snake case columns: %s",
                "\n".join(non_snake_case_cols),
            )
            logger.info("rename to snake case columns")
            df = df.withColumns(transform_dict)
            final_cols = final_cols + list(transform_dict.keys())
            df = df.select(*final_cols)
        return df


class RenameColumns(BaseSparkTransform):
    """Rename Columns Transform model."""

    op: Literal["rename_columns"]
    columns: list[RenameColMap] = Field(
        description="A list of RenameColMap model."
    )
    allow_fill_null_when_col_not_exist: bool = False

    @model_validator(mode="after")
    def __allow_fill_null_rule(self) -> Self:
        """Validate each column model with allow autofill null on not-existing
        column need its data type.
        """
        if self.allow_fill_null_when_col_not_exist:
            if any(c is None for c in self.columns):
                raise ValueError(
                    "`allow_fill_null_when_col_not_exist` property must use "
                    "with `dtype` property"
                )
        return self

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> list[PairCol]:
        """Apply to Rename Column transform.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.
        """
        map_cols: list[PairCol] = []
        for c in self.columns:
            if self.allow_fill_null_when_col_not_exist:
                map_cols.append(c.get_rename_pair_fix_non_existed_by_null(df))
            else:
                map_cols.append(c.get_rename_pair())
        logger.info(f"Rename columns statement:\n{map_cols}")
        return map_cols

    def apply_group(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> PairCol | list[PairCol]:
        return self.apply(df, engine, spark=spark, **kwargs)


class SelectColumns(BaseSparkTransform):
    """Select Columns Operator transform model."""

    op: Literal["select"]
    columns: list[str]
    allow_missing: bool = False

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply to Select Column.

        Args:
            df (DataFrame): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            DataFrame: A column selected Spark DataFrame.
        """
        selection: list[str] = self.columns
        if self.allow_missing:
            selection: list[str] = [c for c in self.columns if c in df.columns]
        logger.info("Select Columns")
        for c in selection:
            logger.info(f"> - {c}")
        return df.select(*selection)


class DropColumns(BaseSparkTransform):
    op: Literal["drop_columns"]
    columns: list[str]
    allow_missing_columns: bool = False

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply drop column.

        Args:
            df (DataFrame): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            DataFrame:
        """
        current_cols = df.columns
        target_cols = self.columns
        if self.allow_missing_columns:
            target_cols = [c for c in self.columns if c in current_cols]
        return df.drop(*target_cols)


class CleanMongoJsonStr(BaseSparkTransform):
    op: Literal["clean_mongo_json_string"]
    source: str
    use_java: bool = True

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> PairCol:
        """Apply to Clean Mongo Json string.

        Args:
            df (DataFrame): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.
        """
        from pyspark.sql.functions import col, expr
        from pyspark.sql.types import StringType

        from ..udf import clean_mongo_json_udf

        func_name: str = "cleanMongoJsonString"
        if self.use_java:
            java_class_name: str = "custom.CleanMongoJsonStringUDF"
            spark.udf.registerJavaFunction(
                func_name, java_class_name, StringType()
            )
            column: Column = expr(f"{func_name}({self.source})").cast("string")
            return column, self.source
        mongo_udf = clean_mongo_json_udf(spark)
        return mongo_udf(col(self.source)), self.source


class JsonStrToStruct(BaseSparkTransform):
    """Json string to Struct Operator transform model."""

    op: Literal["json_to_struct"]
    source: str
    infer_timestamp: bool = True
    timestamp_format: str = "yyyy-MM-dd'T'HH:mm:ss[.SSS]'Z'"
    timestampntz_format: str = "yyyy-MM-dd'T'HH:mm:ss[.SSS]'Z'"
    mode: Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"] = "FAILFAST"
    tmp_dir: str | None = Field(default=None)
    _tmp_dir: str | None = PrivateAttr(default=None)

    def override_tmp_dir(self) -> str:
        rand_str: str = get_random_str_unique(n=12)
        current_date: date = get_dt_now().date()
        self._tmp_dir: str = (
            f"{self.tmp_dir}/json_string_to_struct/{current_date}/{rand_str}"
        )
        return self._tmp_dir

    def post_apply(self, engine: DictData, **kwargs):
        """Remove temp dir after apply this operator."""
        if self._tmp_dir:
            # TODO: need to add post execute func - to clean up temporary
            #   directory
            pass

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply convert JSON string column into new pyspark dataframe.

        Args:
            df (DataFrame): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.
        """
        # NOTE: spark connect, cannot use RDD, so write to temporary instead
        if is_remote_session(spark):
            temp_dir: str = self.override_tmp_dir()
            logger.info(f"Write data to temporary dir: {temp_dir}")
            (
                df.select(self.source)
                .write.format("text")
                .mode("overwrite")
                .save(temp_dir)
            )
            reader = spark.read.format("json").option("inferSchema", True)
            if not self.infer_timestamp:
                return reader.load(temp_dir, mode=self.mode)
            return (
                reader.option("inferTimestamp", self.infer_timestamp)
                .option("timestampNTZFormat", self.timestampntz_format)
                .load(
                    temp_dir,
                    timestampFormat=self.timestamp_format,
                    mode=self.mode,
                )
            )

        # IMPORTANT: Cannot do `c[str(self.source)]` in lambda func, it will
        #   raise error.
        source_str: str = str(self.source)
        rdd_data = df.rdd.map(lambda c: c[source_str])
        reader = spark.read.option("inferSchema", True)
        if self.infer_timestamp:
            return (
                reader.option("inferTimestamp", self.infer_timestamp)
                .option("timestampNTZFormat", self.timestampntz_format)
                .json(
                    rdd_data,
                    timestampFormat=self.timestamp_format,
                    mode=self.mode,
                )
            )
        return reader.json(rdd_data, mode=self.mode)


class Scd2(BaseSparkTransform):
    """SCD Type 2 Transform operator model."""

    op: Literal["scd2"]
    merge_key: list[str]
    update_key: str
    create_key: str
    col_start_name: str = "_scd2_start_time"
    col_end_name: str = "_scd2_end_time"

    @field_validator(
        "merge_key", mode="before", json_schema_input_type=list[str] | str
    )
    def __convert_merge_key_to_dict(cls, value: Any) -> Any:
        """Convert the `merge_key` column that pass with string type to list."""
        return [value] if isinstance(value, str) else value

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply to SCD Type 2.

        This method do 4 steps:
            1: Get the target DataFrame from data warehouse (if exists)
            2: Add `_scd2_start_time` and `_scd_end_time` column to the source
                DataFrame.
                - `_scd2_start_time`:
                    - if id of the row already exist in target table,
                      _scd2_start_time will be update_key for that row
                    - if id of the row doesn't exist in target table (new record),
                      _scd2_start_time will be create_key for that row
                - `_scd2_end_time`: Defaults to NULL
            3: Update df_target._scd_end_time from null to df_source._scd2_start_time in some rows
            4: Combine both DataFrames from the step 2 and 3 together and send
                it to sink process.

        Args:
            df (DataFrame): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            Dataframe: the final dataframe that ready to write to the table
        """
        from pyspark.sql.functions import col, lit

        db_name: str | None = None
        table_name: str | None = None

        # NOTE: Check if config_dict is not empty (not {})
        _sink: Sink = engine["engine"].sink
        if _sink.type in ("iceberg", "hdfs"):
            logger.info("Use sink's configuration from config dict")
            db_name: str = _sink.database
            table_name: str = _sink.table_name
        elif _sink.type.startswith("local"):
            raise NotImplementedError("Does not support local sink type.")
        else:
            logger.info(
                "No `db_name` and `table_name` provided. so no need to connect to "
                "target table and just add the _scd2_start_time and "
                "_scd2_end_time to source df"
            )
        if (
            db_name
            and table_name
            and is_table_exist(
                spark=spark, database=db_name, table_name=table_name
            )
        ):
            # Step01: get df_tgt
            tgt_df = spark.sql(f"SELECT * FROM {db_name}.{table_name}").filter(
                col(self.col_end_name).isNull()
            )
            # Step02: add _scd2_start_time and _scd_end_time column to df_source
            src_df = self._add_start_and_end_in_src(tgt_df=tgt_df, src_df=df)

            # Step03: update df_tgt._scd_end_time from null to df_source._scd2_start_time
            rs_df = self._update_end_time_in_tgt(tgt_df=tgt_df, src_df=src_df)

            # Step04: combine 2 df together
            return rs_df.unionByName(src_df, allowMissingColumns=True)

        # NOTE: if target table doesn't exist or db_name and table name are not
        #   provided, choose created_at as {scd2_start_col}.
        return df.withColumns(
            {
                self.col_start_name: col(self.create_key),
                self.col_end_name: lit(None).cast("timestamp"),
            }
        )

    def _add_start_and_end_in_src(
        self,
        tgt_df: DataFrame,
        src_df: DataFrame,
    ) -> DataFrame:
        """Add thr `_scd2_start_time` and `_scd_end_time` columns to the source
        DataFrame.
        """
        from pyspark.sql import functions as f

        # NOTE: get existing id in target table and max updated_at value
        tgt_df_max: DataFrame = tgt_df.groupby(self.merge_key).agg(
            f.max(self.update_key).alias("_max_update_key")
        )
        src_df_merge = src_df.join(tgt_df_max, on=self.merge_key, how="left")

        # NOTE: Persist it before checking the wrong updated_at value to prevent
        #   recompute data.
        src_df_merge.persist()

        # NOTE: Check the wrong `updated_at` value from source
        #   - if target['_max_update_key'] > source[update_key]: it should not
        #       happen, we cannot get the past data from source.
        wrong_df: DataFrame = src_df_merge.filter(
            f.col("_max_update_key") > f.col(self.update_key)
        ).select(*(self.merge_key + [self.update_key, "_max_update_key"]))

        # NOTE: it's already persisted so should not take that much time to
        #   execute take(1)
        if not wrong_df.isEmpty():
            wrong_data: list[Row] = wrong_df.take(1)
            raise ValueError(
                f"There are wrong `{self.update_key}` from source. It's lower "
                f"than the existing value in the target table.",
                f"This is the first row of the wrong data -> {wrong_data[0]}",
            )

        # NOTE: if target['_max_update_key'] = source[update_key]: this case
        #   will happen when that rows already exist in the target because of
        #   rerunning the same job. So no need to reprocess and save it again.
        src_df_merge_filter = src_df_merge.filter(
            (tgt_df_max["_max_update_key"] < src_df[self.update_key])
            | (tgt_df_max["_max_update_key"].isNull())
        )

        if src_df_merge_filter.isEmpty():
            logger.info(
                "Every row in source df already existed in the target table. "
                "So you will get an empty dataframe result"
            )

        # NOTE: create scd2_start_col by checking join result
        check_merge_not_null: list[Column] = [
            tgt_df_max[key].isNotNull() for key in self.merge_key
        ]
        final_df: DataFrame = src_df_merge_filter.withColumns(
            {
                self.col_start_name: (
                    f.when(
                        *check_merge_not_null, src_df[self.update_key]
                    ).otherwise(src_df[self.create_key])
                ),
                self.col_end_name: f.lit(None).cast("timestamp"),
            }
        )

        # NOTE: remove merge key and _max_update_key that getting from join
        return final_df.select(
            src_df["*"], self.col_start_name, self.col_end_name
        )

    def _update_end_time_in_tgt(
        self,
        tgt_df: DataFrame,
        src_df: DataFrame,
    ) -> DataFrame:
        """Update tgt_df._scd_end_time from null to src_df._scd2_start_time."""
        merge_condition = [
            tgt_df[key] == src_df[key] for key in self.merge_key
        ] + [tgt_df[self.update_key] < src_df[self.update_key]]

        # NOTE: select only necessary using withColumn because we have
        #   duplicated column names
        final_df: DataFrame = tgt_df.join(
            src_df,
            on=merge_condition,
            how="inner",
        ).select(tgt_df["*"], src_df[self.col_start_name])

        # NOTE: update tgt_df[scd2_end_col] = src_df[scd2_start_col], then keep o
        #   nly target and new `_scd_end_time` column
        return final_df.withColumn(
            self.col_end_name, src_df[self.col_start_name]
        ).drop(src_df[self.col_start_name])


class ExplodeArrayColumn(BaseSparkTransform):
    """Explode Array Column Operation transform model."""

    op: Literal["explode_array"]
    explode_col: str
    is_explode_outer: bool = True
    is_return_position: bool = False
    position_prefix_name: str = "_index_pos"

    @field_validator("explode_col", "position_prefix_name", mode="after")
    def __validate_explode_col(cls, value: str) -> str:
        if "." in value:
            raise ValueError("Do not pass `.`, it supports only first level.")
        return value

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply to Explode Array Column.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.
        """
        from pyspark.sql import functions as f
        from pyspark.sql.functions import (
            explode,
            explode_outer,
            posexplode,
            posexplode_outer,
        )

        if not self.is_return_position:
            logger.info("Start Explode Array Column")
            if self.is_explode_outer:
                return df.withColumn(
                    self.explode_col, explode_outer(f.col(self.explode_col))
                )
            return df.withColumn(
                self.explode_col, explode(f.col(self.explode_col))
            )

        logger.info("Start Explode Array Column with Position.")
        pos_col_name: str = f"{self.position_prefix_name}_{self.explode_col}"
        if df.schema[self.explode_col].dataType.typeName() not in ("array",):
            raise ValueError("Support only ('array', ) data type")

        if pos_col_name in df.schema.fieldNames():
            raise ValueError(
                "Position column name is duplicated, please reconfigure "
                "`position_prefix_name` field."
            )
        selection: list[str] = [c for c in df.columns if c != self.explode_col]
        if self.is_explode_outer:
            return df.select(
                *selection,
                posexplode_outer(self.explode_col).alias(
                    pos_col_name, self.explode_col
                ),
            )
        return df.select(
            *selection,
            posexplode(self.explode_col).alias(pos_col_name, self.explode_col),
        )


class FlattenAllExceptArray(BaseSparkTransform):
    """Flatten all Columns except Array datatype Operator transform model."""

    op: Literal["flatten_all_columns_except_array"]

    def apply(
        self,
        df: DataFrame,
        engine: DictData,
        *,
        spark: SparkSession | None = None,
        **kwargs,
    ) -> DataFrame:
        """Apply to Flatten all Columns except Array data type.

        Args:
            df (Any): A Spark DataFrame.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            spark (SparkSession, default None): A Spark session.

        Returns:
            DataFrame:
        """
        from pyspark.sql.functions import col

        transform_dict: dict[str, Column] = {}
        for c in extract_cols_without_array(schema=df.schema):
            flatten_col: str = "_".join(c.split("."))
            transform_dict[flatten_col] = col(c)

        logger.info("Start Flatten all columns except array")
        for k, v in transform_dict.items():
            logger.info(f"> Target col: {k}, from: {v}")
        return df.withColumns(transform_dict).select(
            *list(transform_dict.keys())
        )



from typing import Annotated, Union

from pydantic import Field

from .console import Console
from .empty import Empty
from .iceberg.sink import Iceberg

Sink = Annotated[
    Union[
        Console,
        Empty,
        Iceberg,
    ],
    Field(discriminator="type"),
]



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from pydantic import Field

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession

from .....__types import DictData
from .....models import Shape
from .....utils import bool2str
from ....__abc import BaseSource

logger = logging.getLogger("jett")


class LocalCSVFile(BaseSource):
    """Local CSV file data source."""

    type: Literal["local"]
    file_format: Literal["csv"]
    path: str
    delimiter: str = "|"
    header: bool = Field(default=True)
    lazy: bool = True
    sample_records: int | None = 200

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        """Load CSV file to DuckDB Relation object."""
        file_format: str = Path(self.path).suffix
        if file_format not in (".csv",):
            raise NotImplementedError(
                f"Local file format: {file_format!r} does not match with the "
                f"config file format."
            )
        spark: SparkSession = engine["spark"]
        fullpath: str = "file://" + str(Path(self.path).resolve().absolute())
        logger.info(f"🚰 Source - Start Load: {fullpath}")

        df: DataFrame = (
            spark.read.format(self.file_format)
            .option("header", "true" if self.header else "false")
            .option("delimiter", self.delimiter)
            .option("inferSchema", "true")
            .load(fullpath)
        )
        df.show(10, truncate=False)
        shape: tuple[int, int] = (
            0 if self.lazy else df.count(),
            len(df.columns),
        )
        return df, Shape.from_tuple(shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path


class LocalJsonFile(BaseSource):
    """Local JSON file data source."""

    type: Literal["local"]
    file_format: Literal["json"]
    path: str
    multiline: bool = False
    lazy: bool = True

    def load(self, engine: DictData, **kwargs) -> tuple[DataFrame, Shape]:
        file_format: str = Path(self.path).suffix
        if file_format not in (".json", ".bjson"):
            raise NotImplementedError(
                f"Local file format: {file_format!r} does not match with the "
                f"config file format."
            )
        spark: SparkSession = engine["spark"]
        fullpath: str = "file://" + str(Path(self.path).resolve().absolute())
        logger.info(f"🚰 Source - Start Load: {fullpath}")
        df: DataFrame = (
            spark.read.format(self.file_format)
            .option("inferSchema", "true")
            .option("multiline", bool2str(self.multiline))
            .load(fullpath)
        )
        df.show(10, truncate=False)
        shape: tuple[int, int] = (
            0 if self.lazy else df.count(),
            len(df.columns),
        )
        return df, Shape.from_tuple(shape)

    def inlet(self) -> tuple[str, str]:
        return "local", self.path



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from daft import DataFrame, Schema

from ....__types import DictData
from ....models import MetricOperatorOrder
from ...__abc import BaseTransform


class BaseDaftTransform(BaseTransform, ABC):

    @abstractmethod
    def apply(
        self,
        df: "DataFrame",
        engine: DictData,
        metric: MetricOperatorOrder,
        **kwargs,
    ) -> "DataFrame":
        """Apply Transform Operation on the Daft DataFrame object."""

    @staticmethod
    def sync_schema(
        pre: "Schema", post: "Schema", metric, **kwargs
    ) -> None: ...



from __future__ import annotations

import logging
import subprocess
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Literal, TypedDict

from pydantic import Field
from pydantic.functional_validators import model_validator
from typing_extensions import Self

from ...__about__ import __version__
from ...__types import DictData, PrimitiveType
from ...models import (
    ColDetail,
    Context,
    MetricEngine,
    MetricOperatorTransform,
    MetricTransform,
    Result,
)
from ...utils import exec_command, regex_by_group, sort_non_sensitive_str
from ..__abc import BaseEngine
from .schema_change import (
    clean_col_except_item_from_extract_array,
)
from .sink import Sink
from .source import Source
from .temp_storage import TempStorage
from .transform import Transform
from .utils import (
    add_spark_cmd,
    clean_tz_for_extra_java_options,
    extract_cols_without_array,
    extract_spark_conf_keys,
    is_remote_session,
    schema2dict,
    yarn_fetch_log,
    yarn_kill,
)

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrame, SparkSession
    from pyspark.sql.connect.session import DataFrame as DataFrameRemote
    from pyspark.sql.connect.session import SparkSession as SparkRemoteSession
    from pyspark.sql.types import StructType

    PairCol = tuple[Column, str]
    AnyDataFrame = DataFrame | DataFrameRemote
    AnySparkSession = SparkSession | SparkRemoteSession

logger = logging.getLogger("jett")


class EngineContext(TypedDict):
    """Engine Context dict typed for Spark engine execution."""

    spark: SparkSession
    engine: Spark
    temp_storage: dict[str, Any]


class Spark(BaseEngine):
    """Spark Engine model. This engine will execute Pyspark via connect on the
    current context.
    """

    type: Literal["spark"] = Field(description="A type of spark engine.")
    app_name: str | None = Field(
        default=None, description="An application name."
    )
    enable_collect_result: bool = Field(
        default=False,
        description=(
            "A flag that use for enable collect data from LazyFrame after the "
            "execution done."
        ),
    )
    source: Source = Field(description="A source model.")
    sink: Sink = Field(description="A sink model.")
    transforms: list[Transform] = Field(
        default_factory=list,
        description="A list of transform model.",
    )
    master: str | None = None
    remote: str | None = None
    deploy_mode: Literal["cluster"] = "cluster"
    jars: list[str] = Field(default_factory=list)
    files: list[str] = Field(default_factory=list)
    timezone: str = "UTC"
    driver_memory: str = "1G"
    driver_cores: int = 1
    executor_memory: str = "2G"
    executor_cores: int = 1
    num_executors: int = 2
    max_executors: int = 5
    show_all_spark_submit_log: bool = False
    enable_hive_support: bool = False
    conf: dict[str, PrimitiveType | None] = Field(default_factory=dict)

    def _validate_master_and_remote(self) -> None:
        """Validate master and remote fields.

        Raises:
            ValueError: If the master and remote fields does not set, or it set
                together.
        """
        if (self.master and self.remote) or (
            self.master is None and self.remote is None
        ):
            raise ValueError(
                "please choose either master or remote to be specify."
            )

    @model_validator(mode="after")
    def validate_after(self) -> Self:
        """Validate Spark engine after mode."""
        self._validate_master_and_remote()
        return self

    def session(self, context: Context, **kwargs) -> SparkSession:
        """Create Spark Session with this model engine configuration data.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.

        Returns:
            SparkSession: return spark session.
        """
        from pyspark.sql import SparkSession

        _ = kwargs
        builder = SparkSession.builder
        if self.master:
            builder = builder.master(self.master)
        else:
            builder = builder.remote(self.remote)

        builder = builder.appName(self.app_name or self.name).config(
            map={
                "spark.driver.memory": "2G",
                "spark.driver.cores": 1,
            }
            | self.conf
        )
        if self.jars:
            builder = builder.config("spark.jars", ",".join(self.jars))

        if self.files:
            builder = builder.config("spark.files", ",".join(self.files))

        if self.enable_hive_support:
            builder = builder.enableHiveSupport()

        spark: SparkSession = builder.getOrCreate()

        logger.info(f"Spark Session: {spark}")

        if is_remote_session(spark):
            context["metric_engine"].app_id = spark.sparkContext.applicationId
        return spark

    def set_engine_context(self, context: Context, **kwargs) -> EngineContext:
        """Create Spark Engine Context data for passing to the execute method.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.

        Returns:
            DictData: A mapping of necessary data for Spark execution.
        """
        return {
            "spark": self.session(context, **kwargs),
            "engine": self,
            "temp_storage": {},
        }

    def execute(
        self, context: Context, engine: DictData, metric: MetricEngine
    ) -> DataFrame:
        """Execute Spark engine method. This method do procedure process to load
        data from source and save it to the sink model.

            This method start process with 3 steps:
                1. Handle source load method
                2. Handle apply operator transform
                3. Handle sink save method

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricEngine): A metric engine that was set from handler
                step for passing custom metric data.

        Returns:
            DataFrame: A result DataFrame API.
        """
        logger.info("🏗️ Start execute with Spark engine.")
        # NOTE: Start run source handler.
        df: DataFrame = self.source.handle_load(context, engine=engine)

        # NOTE: Start run transform handler.
        df: DataFrame = self.handle_apply(df, context, engine=engine)

        # NOTE: Start run sink handler.
        self.sink.handle_save(df, context, engine=engine)
        return df

    def set_result(self, df: DataFrame, context: DictData) -> Result:
        """Set the Result object for this Spark engine.

        Args:
            df (DataFrame): A Spark DataFrame.
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.

        Returns:
            Result: A result object that catch schema and sample data from the
            execution result DataFrame.
        """
        if self.enable_collect_result:
            logger.warning(
                "⚠️ If you collect results from the DataFrame, it will fetch all "
                "records in the DataFrame. This is not ideal for any ETL "
                "pipeline with large data sizes. Please use it only for "
                "testing purposes or with smaller datasets."
            )
        schema: StructType = df.schema
        return Result(
            data=df.collect() if self.enable_collect_result else [],
            columns=[
                ColDetail(name=f.name, dtype=f.dataType.simpleString())
                for f in schema
            ],
            schema_dict=df.schema.jsonValue(),
        )

    def post_execute(
        self,
        context: Context,
        *,
        engine: DictData,
        exception: Exception | None = None,
        **kwargs,
    ) -> None:
        """Post-execute Spark engine. This method use to clear a temp storage
        after execute all layer of execution method.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            exception (Exception, default None): An exception class that raise
                from execution method. It will be None value if it completes
                without raise any exception.
        """
        logger.info("🎢 Start Post Spark Execution ...")
        tmp_st: dict[str, TempStorage | None] = engine["temp_storage"]

        # NOTE: Clear source temp storage.
        temp_source: TempStorage | None = tmp_st.get("source")
        if temp_source is not None and temp_source.is_written:
            temp_source.teardown()

        # NOTE: Clear sink temp storage.
        temp_sink: TempStorage | None = tmp_st.get("sink")
        if temp_sink is not None and temp_sink.is_written:
            temp_sink.teardown()

        engine["spark"].stop()

    def apply(
        self,
        df: DataFrame,
        context: Context,
        engine: DictData,
        metric: MetricTransform,
        **kwargs,
    ) -> DataFrame:
        """Apply Spark engine transformation to the source. This method will
        apply all operators by priority.

            The Spark engine apply the split sub-apply on transform operators
        to 3 layers of priority:
            1. pre layer: An ordering transform before group transform
            2. group layer: A group transform
            3. post layer: A ordering transform after group transform

        Args:
            df (DataFrame): A Spark DataFrame.
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            engine (DictData): An engine context data that was created from the
                `post_execute` method. That will contain engine model, engine
                session object for this execution, or it can be specific config
                that was generated on that current execution.
            metric (MetricTransform): A metric transform that was set from
                handler step for passing custom metric data.
        """
        spark: SparkSession = engine["spark"]
        priority: list[Transform] = []
        groups: list[Transform] = []
        fallback: list[Transform] = []

        logger.debug("⚙️ Transform - Prepare transform analyzer ...")
        for t in self.transforms:
            if t.priority == "pre":
                priority.append(t)
            elif t.priority == "group":
                groups.append(t)
            else:
                fallback.append(t)

        logger.info(f"⚙️ Priority transform count: {len(priority)}")
        for t in priority:
            logger.info(f"Start priority operator: {t.op!r}")
            df: DataFrame = t.handle_apply(
                df,
                context=context,
                engine=engine,
                spark=spark,
            )

        logger.info(f"⚙️ Group transform count: {len(groups)}")
        if groups:
            maps: dict[str, Column] = {}
            for t in groups:
                # NOTE: Generate sub transform for apply group transform.
                maps.update(
                    t.handle_apply_group(
                        df, context, engine=engine, spark=spark
                    )
                )

            if len(maps) > 0:
                for k, v in maps.items():
                    logger.info(f"... target col: {k}, from: {v}")
                pre_schema: StructType = df.schema
                df: DataFrame = df.withColumns(maps)
                self.sync_schema_group(
                    pre_schema, df.schema, context=context, spark=spark
                )

        logger.info(f"⚙️ Fallback transform count: {len(fallback)}")
        for t in fallback:
            logger.info(f"Start fallback operator: {t.op!r}")
            df: DataFrame = t.handle_apply(
                df,
                context,
                engine=engine,
                spark=spark,
            )

        return df

    @staticmethod
    def sync_schema_group(
        pre: StructType,
        post: StructType,
        *,
        context: Context,
        spark: SparkSession | None = None,
    ) -> None:
        """Update pre- and post-schema metrics data for any transform that apply
        to the source DataFrame API until it save to the sink.

        Args:
            pre (StructType): A pre-transform StructType object.
            post (StructType): A post-transform StructType object.
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
            spark (SparkSession, default None): A Spark session.
        """
        pre_schema = spark.createDataFrame(data=[], schema=pre).schema
        post_schema = spark.createDataFrame(data=[], schema=post).schema
        pre_no_array = sort_non_sensitive_str(
            extract_cols_without_array(schema=pre_schema)
        )
        post_no_array = sort_non_sensitive_str(
            extract_cols_without_array(schema=post_schema)
        )
        metric: MetricOperatorTransform = context["metric_group_transform"]

        # NOTE: Start update the pre- and post-schema metric.
        metric.transform_pre = {
            "schema": schema2dict(pre, sorted_by_name=True),
            "schema_no_array": pre_no_array,
        }
        metric.transform_post = {
            "schema": schema2dict(post, sorted_by_name=True),
            "schema_no_array": post_no_array,
        }
        context["metric_operator"].append(metric)


class SparkSubmit(Spark):
    """Spark Submit Engine model."""

    type: Literal["spark-submit"] = Field(
        description="A type of spark submit engine."
    )
    entrypoint: str = Field(
        description="An entrypoint of the Spark submit command.",
    )
    archives: list[str] = Field(default_factory=list)
    py_files: list[str] = Field(default_factory=list)
    env_var: dict[str, str] = Field(default_factory=dict)
    queue: str | None = None
    keytab_path: str | None = None
    keytab_principal: str | None = None
    yarn_app_id: str | None = Field(
        default=None,
        description="A YARN application ID after submit the Spark application.",
    )

    @model_validator(mode="after")
    def validate_after(self) -> Self:
        """Inherit after validation from Spark engine model for checking master
        field value that allow to use this Spark submit engine.
        """
        self._validate_master_and_remote()
        if self.master and self.master.startswith("local"):
            raise ValueError("Spark Submit does not support for local master.")
        return self

    def get_cmd(self) -> list[str]:
        """Generate Spark submit commands from all model field."""
        cmd: list[str] = ["spark-submit"]
        add_spark_cmd(cmd, "master", self.master)
        add_spark_cmd(cmd, "name", self.app_name)
        add_spark_cmd(cmd, "deploy-mode", self.deploy_mode)
        add_spark_cmd(cmd, "archives", self.archives)
        add_spark_cmd(cmd, "files", self.files)
        add_spark_cmd(cmd, "py-files", self.py_files)
        add_spark_cmd(cmd, "jars", self.jars)
        add_spark_cmd(cmd, "driver-memory", self.driver_memory)
        add_spark_cmd(cmd, "driver-cores", self.driver_cores)
        add_spark_cmd(cmd, "executor-memory", self.executor_memory)
        add_spark_cmd(cmd, "executor-cores", self.executor_cores)
        add_spark_cmd(cmd, "num-executors", self.num_executors)
        add_spark_cmd(cmd, "keytab", self.keytab_path)
        add_spark_cmd(cmd, "principal", self.keytab_principal)
        add_spark_cmd(cmd, "queue", self.queue)

        if self.enable_hive_support:
            cmd.extend(["--conf", '"spark.sql.catalogImplementation=hive"'])

        if self.max_executors:
            cmd.extend(
                [
                    '--conf "spark.dynamicAllocation.enabled=True"',
                    f'--conf "spark.dynamicAllocation.maxExecutors={str(self.max_executors)}"',
                ]
            )

        # NOTE: Set timezone
        driver_extra_ops = clean_tz_for_extra_java_options(
            option=self.conf.get("spark.driver.extraJavaOptions", ""),
            tz=self.timezone,
        )
        executor_extra_ops = clean_tz_for_extra_java_options(
            option=self.conf.get("spark.executor.extraJavaOptions", ""),
            tz=self.timezone,
        )
        cmd.extend(
            [
                "--conf",
                f'"spark.sql.session.timeZone={self.timezone}"',
                "--conf",
                f'"spark.driver.extraJavaOptions={driver_extra_ops}"',
                "--conf",
                f'"spark.executor.extraJavaOptions={executor_extra_ops}"',
            ]
        )

        if self.conf:
            exist_spark_conf: list[str] = extract_spark_conf_keys(cmd=cmd)
            cmd.extend(
                [
                    f'--conf "{k}={self.conf[k]}"'
                    for k in self.conf
                    if k not in exist_spark_conf
                ]
            )

        cmd.extend(["--conf", f'"spark.jett.version={__version__}"'])

        if self.master == "yarn" and self.deploy_mode == "cluster":
            for k, v in self.env_var.items():
                cmd.extend(["--conf", f'"spark.yarn.appMasterEnv.{k}={v}"'])

        cmd.append(self.entrypoint)
        return cmd

    def iter_submit_log(self, logs: Iterator[str]) -> None:
        """Logging spark submit log and grep YARN application ID.

        Args:
            logs (Iterator[str]): A yield of log statement.
        """
        already_log_yarn_url: bool = False
        states: list[str] = []
        for log in logs:
            log = log.strip()
            if self.show_all_spark_submit_log:
                logger.info(log)
            else:
                state: str = regex_by_group(log, regex=r"state:\s+(.+?)\)", n=1)
                if state not in states and state != "":
                    logger.info("application state: %s", state)
                    states.append(state)

            if self.master != "yarn" or self.deploy_mode != "cluster":
                continue

            app_id: str = regex_by_group(log, regex=r"application[0-9_]+")
            if app_id != "":
                self.yarn_app_id = app_id

            if self.show_all_spark_submit_log:
                continue

            if "tracking URL:" in log:
                _yarn_url: str = log.replace("tracking URL:", "").strip()
                if not already_log_yarn_url:
                    logger.info(f"YARN app id: {self.yarn_app_id}")
                    logger.info(f"YARN app tracking url: {_yarn_url}")
                    already_log_yarn_url = True

    def submit(self, context: Context) -> None:
        """Submit Spark application to YARN.

        Args:
            context (Context): A execution context that was created from the
                core operator execution step this context will keep all operator
                metadata and metric data before emit them to metric config
                model.
        """
        submit_cmd: list[str] = self.get_cmd()
        logger.info(f"Spark Submit Command:\n{submit_cmd}")

        proc: subprocess.Popen = exec_command(" ".join(submit_cmd))
        self.iter_submit_log(iter(proc.stdout.readline, ""))
        return_code: int = proc.wait()
        if return_code != 0:
            # NOTE: kill app if YARN application in cluster mode
            if self.master == "yarn" and self.deploy_mode == "cluster":
                yarn_kill(app_id=self.yarn_app_id)
                yarn_fetch_log(
                    application_id=self.yarn_app_id, log_file="stdout,stderr"
                )
            raise RuntimeError(
                f"Failed to execute spark application, stderr: "
                f"{str(proc.stderr)}"
            )
        if self.master == "yarn" and self.deploy_mode == "cluster":
            yarn_fetch_log(application_id=self.yarn_app_id, log_file="stdout")

        context.update(
            {
                "metric_engine": MetricEngine(
                    app_id=self.yarn_app_id or self.app_name,
                )
            }
        )

    def apply(
        self,
        df: DataFrame,
        context: Context,
        *,
        engine: DictData,
        **kwargs,
    ) -> DataFrame:
        """Apply Spark engine transformation to the source. This method will
        apply all operators by priority.

            This engine does not allow to use apply method.
        """
        raise NotImplementedError(
            "Spark Submit Engine does not support apply transform."
        )

    def execute(
        self, context: Context, engine: DictData, metric: MetricEngine
    ) -> DataFrame:
        logger.info("Start execute with Spark Submit engine.")
        raise NotImplementedError(
            "Spark Submit Engine does not support direct execute yet."
        )



class BaseError(Exception):
    """Base Tool Exception class."""


class ToolError(BaseError):
    """Main Tool Exception class."""


class ToolValidationError(ToolError):
    """A Custom Exception raised when validation failed."""


class ToolTransformError(ToolError):
    """A Custom Exception raised when transform failed."""


class UtilError(Exception): ...


class CmdExecError(UtilError):  # pragma: no cov
    """Custom exception raised for command execution failures.

    This exception provides context about the failure, including the command's
    return code and any output from its standard error stream.
    """

    def __init__(self, message: str, return_code: int, stderr: str | None):
        super().__init__(message)
        self.return_code = return_code
        self.stderr = stderr

    def __str__(self):
        message = super().__str__()
        if self.stderr:
            return (
                f"({self.return_code}): {message}\n--- Stderr ---\n"
                f"{self.stderr.strip()}"
            )
        return message
